# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
import io

from dataclasses import dataclass
from typing import Type, List, Dict, Tuple
from volatility3.framework import constants, exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.objects import utility
from volatility3.framework.constants import architectures
from volatility3.framework.symbols import linux

# Image manipulation functions are kept in the plugin,
# to prevent a general exit on missing PIL (pillow) dependency.
try:
    from PIL import Image

    has_pil = True
except ImportError:
    has_pil = False

vollog = logging.getLogger(__name__)


@dataclass
class Framebuffer:
    """Framebuffer object internal representation. This is useful to unify a framebuffer with precalculated
    properties and pass it through functions conveniently."""

    id: str
    xres_virtual: int
    yres_virtual: int
    line_length: int
    bpp: int
    """Bits Per Pixel"""
    size: int
    color_fields: Dict[str, Tuple[int, int, int]]
    fb_info: interfaces.objects.ObjectInterface


class Fbdev(interfaces.plugins.PluginInterface):
    """Extract framebuffers from the fbdev graphics subsystem"""

    _version = (1, 0, 0)
    _required_framework_version = (2, 11, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="linuxutils", component=linux.LinuxUtilities, version=(2, 2, 0)
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Dump framebuffers",
                default=False,
                optional=True,
            ),
        ]

    @classmethod
    def parse_fb_pixel_bitfields(
        cls, fb_var_screeninfo: interfaces.objects.ObjectInterface
    ) -> Dict[str, Tuple[int, int, int]]:
        """Organize a framebuffer pixel format into a dictionary.
        This is needed to know the position and bitlength of a color inside
        a pixel.

        Args:
            fb_var_screeninfo: a fb_var_screeninfo kernel object instance

        Returns:
            The color fields mappings

        Documentation:
            include/uapi/linux/fb.h:
                struct fb_bitfield {
                    __u32 offset;			/* beginning of bitfield	*/
                    __u32 length;			/* length of bitfield		*/
                    __u32 msb_right;		/* != 0 : Most significant bit is right */
                };
        """
        # Naturally order by RGBA
        color_mappings = [
            ("R", fb_var_screeninfo.red),
            ("G", fb_var_screeninfo.green),
            ("B", fb_var_screeninfo.blue),
            ("A", fb_var_screeninfo.transp),
        ]
        color_fields = {}
        for color_code, fb_bitfield in color_mappings:
            color_fields[color_code] = (
                int(fb_bitfield.offset),
                int(fb_bitfield.length),
                int(fb_bitfield.msb_right),
            )
        return color_fields

    @classmethod
    def convert_fb_raw_buffer_to_image(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_name: str,
        fb: Framebuffer,
    ):
        """Convert raw framebuffer pixels to an image.

        Args:
            fb: the relevant Framebuffer object

        Returns:
            A PIL Image object

        Documentation:
            include/uapi/linux/fb.h:
                /* Interpretation of offset for color fields: All offsets are from the right,
                * inside a "pixel" value, which is exactly 'bits_per_pixel' wide (means: you
                * can use the offset as right argument to <<). A pixel afterwards is a bit
                * stream and is written to video memory as that unmodified.
        """
        kernel = context.modules[kernel_name]
        kernel_layer = context.layers[kernel.layer_name]

        raw_pixels = io.BytesIO(kernel_layer.read(fb.fb_info.screen_base, fb.size))
        bytes_per_pixel = fb.bpp // 8
        image = Image.new("RGBA", (fb.xres_virtual, fb.yres_virtual))

        # This is not designed to be extremely fast (numpy isn't available),
        # but convenient and dynamic for any color field layout.
        for y in range(fb.yres_virtual):
            for x in range(fb.xres_virtual):
                raw_pixel = int.from_bytes(raw_pixels.read(bytes_per_pixel), "little")
                pixel = [0, 0, 0, 255]
                # The framebuffer is expected to have been correctly constructed,
                # especially by parse_fb_pixel_bitfields, to get the needed RGBA mappings.
                for i, color_code in enumerate(["R", "G", "B", "A"]):
                    offset, length, msb_right = fb.color_fields[color_code]
                    if length == 0:
                        continue
                    color_value = (raw_pixel >> offset) & (2**length - 1)
                    if msb_right:
                        # Reverse bit order
                        color_value = int(
                            "{:0{length}b}".format(color_value, length=length)[::-1], 2
                        )
                    pixel[i] = color_value
                image.putpixel((x, y), tuple(pixel))

        return image

    @classmethod
    def dump_fb(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_name: str,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
        fb: Framebuffer,
        convert_to_png_image: bool,
    ) -> str:
        """Dump a Framebuffer buffer to disk.

        Args:
            fb: the relevant Framebuffer object
            convert_to_image: a boolean specifying if the buffer should be converted to an image

        Returns:
            The filename of the dumped buffer.
        """
        kernel = context.modules[kernel_name]
        kernel_layer = context.layers[kernel.layer_name]
        id = "N-A" if isinstance(fb.id, renderers.NotAvailableValue) else fb.id
        base_filename = f"{id}_{fb.xres_virtual}x{fb.yres_virtual}_{fb.bpp}bpp"
        if convert_to_png_image:
            image_object = cls.convert_fb_raw_buffer_to_image(context, kernel_name, fb)
            raw_io_output = io.BytesIO()
            image_object.save(raw_io_output, "PNG")
            final_fb_buffer = raw_io_output.getvalue()
            filename = f"{base_filename}.png"
        else:
            final_fb_buffer = kernel_layer.read(fb.fb_info.screen_base, fb.size)
            filename = f"{base_filename}.raw"

        with open_method(filename) as fp:
            fp.write(final_fb_buffer)
            return fp.preferred_filename

    @classmethod
    def parse_fb_info(
        cls,
        fb_info: interfaces.objects.ObjectInterface,
    ) -> Framebuffer:
        """Parse an fb_info struct
        Args:
            fb_info: an fb_info kernel object live instance

        Returns:
            A Framebuffer object

        Documentation:
            https://docs.kernel.org/fb/api.html:
                - struct fb_fix_screeninfo stores device independent unchangeable information about the frame buffer device and the current format.
                Those information can't be directly modified by applications, but can be changed by the driver when an application modifies the format.
                - struct fb_var_screeninfo stores device independent changeable information about a frame buffer device, its current format and video mode,
                as well as other miscellaneous parameters.
        """
        id = utility.array_to_string(fb_info.fix.id) or renderers.NotAvailableValue()
        color_fields = None

        # 0 = color, 1 = grayscale,	>1 = FOURCC
        if fb_info.var.grayscale in [0, 1]:
            color_fields = cls.parse_fb_pixel_bitfields(fb_info.var)

        # There a lot of tricky pixel formats used by drivers and vendors in include/uapi/linux/videodev2.h.
        # As Volatility3 is not a video format converter, it is best to play it safe and let the user parse
        # the raw data manually (with ffmpeg for example).
        elif fb_info.var.grayscale > 1:
            fourcc = linux.LinuxUtilities.convert_fourcc_code(fb_info.var.grayscale)
            warn_msg = f"""Framebuffer "{id}" uses a FOURCC pixel format "{fourcc}" that isn't natively supported.
You can try using ffmpeg to decode the raw buffer. Example usage:
"ffmpeg -pix_fmts" to list supported formats, then
"ffmpeg -f rawvideo -video_size {fb_info.var.xres_virtual}x{fb_info.var.yres_virtual} -i <FILENAME>.raw -pix_fmt <FORMAT> output.png"."""
            vollog.warning(warn_msg)

        # Prefer using the virtual resolution, instead of the visible one.
        # This prevents missing non-visible data stored in the framebuffer.
        fb = Framebuffer(
            id,
            xres_virtual=fb_info.var.xres_virtual,
            yres_virtual=fb_info.var.yres_virtual,
            line_length=fb_info.fix.line_length,
            bpp=fb_info.var.bits_per_pixel,
            size=fb_info.var.yres_virtual * fb_info.fix.line_length,
            color_fields=color_fields,
            fb_info=fb_info,
        )

        return fb

    def _generator(self):
        if not has_pil:
            vollog.error(
                "PIL (pillow) module is required to use this plugin. Please install it manually or through pyproject.toml."
            )
            return

        kernel_name = self.config["kernel"]
        kernel = self.context.modules[kernel_name]

        if not kernel.has_symbol("num_registered_fb"):
            vollog.error(
                '"num_registered_fb" symbol does not exist in the symbol table. This means you are either analyzing an unsupported kernel version,  your symbol table is corrupt, or the fbdev driver is compiled as a kernel module.'
            )
            return

        try:
            num_registered_fb = kernel.object_from_symbol("num_registered_fb")
        except exceptions.SymbolError:
            vollog.error(
                'Creating an object from "num_registered_fb" caused a symbol error. This is a sign that the symbol table is outdated. Please re-generate your symbol table using the latest dwarf2json'
            )
            return

        if num_registered_fb < 1:
            vollog.info("No registered framebuffer in the fbdev API.")
            return

        registered_fb = kernel.object_from_symbol("registered_fb")
        fb_info_list = utility.array_of_pointers(
            registered_fb,
            num_registered_fb,
            kernel.symbol_table_name + constants.BANG + "fb_info",
            self.context,
        )

        for fb_info in fb_info_list:
            fb = self.parse_fb_info(fb_info)
            file_output = "Disabled"
            if self.config["dump"]:
                try:
                    file_output = self.dump_fb(
                        self.context, kernel_name, self.open, fb, bool(fb.color_fields)
                    )
                    file_output = str(file_output)
                except exceptions.InvalidAddressException as excp:
                    vollog.error(
                        f'Layer {excp.layer_name} failed to read address {hex(excp.invalid_address)} when dumping framebuffer "{fb.id}".'
                    )
                    file_output = renderers.UnreadableValue()

            try:
                fb_device_name = utility.pointer_to_string(
                    fb.fb_info.dev.kobj.name, 256
                )
            except exceptions.InvalidAddressException:
                fb_device_name = renderers.NotAvailableValue()

            yield (
                0,
                (
                    format_hints.Hex(fb.fb_info.screen_base),
                    fb_device_name,
                    fb.id,
                    fb.size,
                    f"{fb.xres_virtual}x{fb.yres_virtual}",
                    fb.bpp,
                    "RUNNING" if fb.fb_info.state == 0 else "SUSPENDED",
                    file_output,
                ),
            )

    def run(self):
        columns = [
            ("Address", format_hints.Hex),
            ("Device", str),
            ("ID", str),
            ("Size", int),
            ("Virtual resolution", str),
            ("BPP", int),
            ("State", str),
            ("Filename", str),
        ]

        return renderers.TreeGrid(
            columns,
            self._generator(),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging

from collections import namedtuple
from typing import List, Tuple, Optional, Generator, Callable

from volatility3.framework.objects import utility
from volatility3.framework import interfaces, renderers, symbols, exceptions
from volatility3.framework.configuration import requirements
from volatility3.plugins import yarascan
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist

vollog = logging.getLogger(__name__)

try:
    import capstone

    has_capstone = True
except ImportError:
    has_capstone = False

# Full details on the techniques used in these plugins to detect EDR-evading malware
# can be found in our 20 page whitepaper submitted to DEFCON along with the presentation
# https://www.volexity.com/wp-content/uploads/2024/08/Defcon24_EDR_Evasion_Detection_White-Paper_Andrew-Case.pdf

syscall_finder_type = namedtuple(
    "syscall_finder_type",
    [
        "get_syscall_target_address",
        "wants_syscall_inst",
        "rule_str",
        "invalid_ops",
        "termination_ops",
    ],
)

syscall_finder_type.__doc__ = """
This type is used to specify how malicious system call invocations should be found.

`get_syscall_target_address` is optionally used to extract the address containing the malicious 'syscall' instruction
`wants_syscall_inst` whether or not this method expects the 'syscall' instruction directly within the malicious code block
`rule` the opcode string to search for the malicious syscall instructions
`invalid_ops` instructions that only appear in invalid code blocks. Stops processing of the code block when encountered.
`termination_ops` instructions that are expected to be present in the code block and that stop processing
"""


class DirectSystemCalls(interfaces.plugins.PluginInterface):
    """Detects the Direct System Call technique used to bypass EDRs"""

    _required_framework_version = (2, 4, 0)

    # 2.0.0 - changes signature of `get_tasks_to_scan`
    _version = (2, 0, 0)

    # DLLs that are expected to host system call invocations
    valid_syscall_handlers = ("ntdll.dll", "win32u.dll")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.syscall_finder = syscall_finder_type(
            # for direct system calls, we find the `syscall` instruction directly, so we already know the address
            None,
            # yes, we want the syscall instruction present as it is what this technique looks for
            True,
            # regex to find "\x0f\x05" (syscall) followed later by "\xc3" (ret)
            # we allow spacing in between to break naive anti-analysis forms (e.g., TarTarus Gate)
            # Standard techniques, such as HellsGate, look like:
            # mov r10, rcx
            # mov eax, <system call number>
            # syscall
            # ret
            "/\\x0f\\x05[^\\xc3]{,24}\\xc3/",
            # any of these will not be in a workable, malicious direct system call block
            ["jmp", "call", "leave", "int3"],
            # the expected form is to end with a "ret" back to the calling code
            ["ret"],
        )

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # create a list of requirements for vadyarascan
        vadyarascan_requirements = [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="yarascanner", component=yarascan.YaraScanner, version=(2, 1, 0)
            ),
            requirements.VersionRequirement(
                name="yarascan", component=yarascan.YaraScan, version=(2, 0, 0)
            ),
        ]

        # get base yarascan requirements for command line options
        yarascan_requirements = yarascan.YaraScan.get_yarascan_option_requirements()

        # return the combined requirements
        return yarascan_requirements + vadyarascan_requirements

    @staticmethod
    def _is_syscall_block(
        disasm_func: Callable,
        syscall_finder: syscall_finder_type,
        data: bytes,
        address: int,
    ) -> Optional[Tuple[str, "capstone._cs_insn"]]:
        """
        Determines if the bytes starting at `data` represent a valid syscall instruction invocation block

        To maliciously invoke the system call instruction, malware must do each of the following:

        1) update RAX to the system call number
        2) update R10 to the first parameter
        3) hit the 'termination' instruction set in `syscall_finder_type`

        We also track whether the 'syscall' instruction was encountered while parsing

        This function is reusable for every technique we found and studied during the DEFCON research timeframe

        Args:
            disasm_func: capstone disassembly function gathered from `get_disasm_function`
            syscall_finder: the method and constraints on the malicious system call blocks that the calling plugin knows how to find
            data: the bytes from memory to search for malicious syscall invocations
            address: the address from where `data` came from in the particular process
        Returns:
            Optional[Tuple[str, capstone._cs_insn]]: For valid blocks, the disassembled bytes in string from and the last (termination) instruction
        """
        found_movr10 = False
        found_movreax = False
        found_syscall = False
        found_end = False
        end_inst = None

        disasm_bytes = ""

        for inst in disasm_func(data, address):
            disasm_bytes += f"{inst.address:#x}: {inst.mnemonic} {inst.op_str}; "

            # an instruction of all 0x00 opcodes
            if inst.opcode.count(0) == len(inst.opcode):
                break

            op = inst.mnemonic

            # invalid op, bail
            if op in syscall_finder.invalid_ops:
                break

            # found the end instruction wanted by the caller
            elif op in syscall_finder.termination_ops:
                found_end = True
                end_inst = inst
                break

            # track this no matter what to make code more re-usable
            elif op == "syscall":
                found_syscall = True

                # if we hit a 'syscall' but RAX or R10 haven't been touched
                # then we are in an invalid path, so bail
                if not syscall_finder.wants_syscall_inst or (
                    not (found_movr10 and found_movreax)
                ):
                    break

            else:
                # attempt to see if any other instruction type wrote to registers
                try:
                    _, regs_written = inst.regs_access()
                except capstone.CsError:
                    continue

                if regs_written:
                    for r in regs_written:
                        # track writes to eax/rax or R10
                        reg = inst.reg_name(r)
                        if reg in ["eax", "rax"]:
                            found_movreax = True

                        elif reg == "r10":
                            found_movr10 = True

        # if any of these are missing, the block is invalid regardless of
        # the technique we are trying to detect now or in the future
        if not (found_movr10 and found_movreax and found_end):
            return None

        # if the finder requires a 'syscall' instruction then bail now if we didn't find one
        if syscall_finder.wants_syscall_inst and not found_syscall:
            return None

        return disasm_bytes, end_inst

    @classmethod
    def get_disasm_function(cls, architecture: str) -> Callable:
        """
        Returns the disassembly handler for the given architecture
        .detail is used to get full instruction information

        Args:
            architecture: the name of the architecture for the process being disassembled
        Returns:
            The disasm function from capstone for the given architecture
        """
        disasm_types = {
            "intel": capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_32),
            "intel64": capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_64),
        }

        disasm_type = disasm_types[architecture]
        disasm_type.detail = True
        return disasm_type.disasm

    @classmethod
    def _is_valid_syscall(
        cls,
        syscall_finder: syscall_finder_type,
        proc_layer: interfaces.layers.DataLayerInterface,
        architecture: str,
        vads: List[Tuple[int, int, str]],
        address: int,
    ) -> Optional[Tuple[int, str]]:
        """
        Args:
            syscall_finder:
            proc_layer: the memory layer of the process being scanned
            architecture: the name of the architecture for the process being disassembled
            vads: the ranges of this process under 10MB
            address: the starting address to check for malicious syscall code blocks

        Returns:
            Optional[Tuple[int, str]]: For valid code blocks, the starting address of the block and the disassembly string
        """
        # the number bytes behind the yara rule hit to scan
        behind = 32

        address = address - behind

        try:
            data = proc_layer.read(address, behind * 2)
        except exceptions.InvalidAddressException:
            return None

        disasm_func = cls.get_disasm_function(architecture)

        # since Intel does not have fixed-size instructions, we have to scan
        # each byte offset and re-disassemble the remaining block
        for offset in range(behind):
            # if this looks like a system call back (r10, rax, ret/jmp)
            syscall_info = cls._is_syscall_block(
                disasm_func, syscall_finder, data[offset:], address + offset
            )
            if syscall_info:
                disasm_bytes, end_inst = syscall_info

                # if we can recover (and require) a target address for this malware technique
                if syscall_finder.get_syscall_target_address:
                    target_address = syscall_finder.get_syscall_target_address(
                        proc_layer, end_inst
                    )

                    # could not determine the address -> invalid basic block
                    if not target_address:
                        continue

                    # we only care about calls to system call DLLs
                    path = cls.get_range_path(vads, target_address)
                    if not isinstance(path, str) or not path.lower().endswith(
                        cls.valid_syscall_handlers
                    ):
                        continue

                # return the address and disassembly string if all checks pass
                return address + offset, disasm_bytes

        return None

    @classmethod
    def get_vad_maps(
        cls,
        task: interfaces.objects.ObjectInterface,
    ) -> List[Tuple[int, int, str]]:
        """Creates a map of start/end addresses within a virtual address
        descriptor tree.

        Args:
            task: The EPROCESS object of which to traverse the vad tree

        Returns:
            An iterable of tuples containing start and end addresses for each descriptor
        """
        vads: List[Tuple[int, int, str]] = []

        # scan regions under 10MB
        scan_max = 10 * 1000 * 1000

        vad_root = task.get_vad_root()

        for vad in vad_root.traverse():
            if vad.get_size() < scan_max:
                vads.append((vad.get_start(), vad.get_size(), vad.get_file_name()))

        return vads

    @classmethod
    def get_range_path(
        cls, ranges: List[Tuple[int, int, str]], address: int
    ) -> Optional[str]:
        """
        Returns the path for the range holding `address`, if found

        Args:
            ranges: VADs collected from `get_vad_maps`
            address: the address to find
        Returns:
            The path holding the address, if any
        """
        for start, size, path in ranges:
            if start <= address < start + size:
                return path

        return None

    @classmethod
    def get_tasks_to_scan(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> Generator[
        Tuple[interfaces.objects.ObjectInterface, str, str, str], None, None
    ]:
        """
        Gathers active processes with the extra information needed
        to detect malicious syscall instructions

        Returns:
            Generator of the process object, name, memory layer, and architecture
        """

        # gather active processes
        filter_func = pslist.PsList.create_active_process_filter()

        kernel = context.modules[kernel_module_name]

        is_32bit_arch = not symbols.symbol_table_is_64bit(
            context=context, symbol_table_name=kernel.symbol_table_name
        )

        for proc in pslist.PsList.list_processes(
            context=context,
            kernel_module_name=kernel_module_name,
            filter_func=filter_func,
        ):
            proc_name = utility.array_to_string(proc.ImageFileName)

            # skip Defender
            if proc_name in ["MsMpEng.exe"]:
                continue

            try:
                proc_layer_name = proc.add_process_layer()
            except exceptions.InvalidAddressException:
                continue

            if is_32bit_arch or proc.get_is_wow64():
                architecture = "intel"
            else:
                architecture = "intel64"

            yield proc, proc_name, proc_layer_name, architecture

    @classmethod
    def _get_rule_hits(
        cls,
        context: interfaces.objects.ObjectInterface,
        proc_layer: interfaces.layers.DataLayerInterface,
        vads: List[Tuple[int, int, str]],
        pattern: str,
    ) -> Generator[Tuple[int, Optional[str]], None, None]:
        """
        Runs the given opcode rule through Yara and returns the address and file path of hits

        Args:
            context:
            proc_layer: the layer to scan
            vads: the ranges inside of the process being scanned
            pattern: the opcodes rule from the plugin to detect a particular EDR-bypass technique

        Returns:
            Generator of the address and file path of hits
        """
        sections = [(vad[0], vad[1]) for vad in vads]

        rule = yarascan.YaraScanner.get_rule(pattern)

        for hit in proc_layer.scan(
            context=context,
            scanner=yarascan.YaraScanner(rules=rule),
            sections=sections,
        ):
            address = hit[0]

            path = cls.get_range_path(vads, address)

            # ignore hits in the system call DLLs
            if isinstance(path, str) and path.lower().endswith(
                cls.valid_syscall_handlers
            ):
                continue

            yield address, path

    def _generator(
        self,
    ) -> Generator[Tuple[int, Tuple[str, int, Optional[str], int, str]], None, None]:
        if not has_capstone:
            vollog.warning(
                "capstone is not installed. This plugin requires capstone to operate."
            )
            return

        for proc, proc_name, proc_layer_name, architecture in self.get_tasks_to_scan(
            self.context, self.config["kernel"]
        ):
            proc_layer = self.context.layers[proc_layer_name]

            vads = self.get_vad_maps(proc)
            if not vads:
                continue

            # for each valid process, look for malicious syscall invocations
            for address, vad_path in self._get_rule_hits(
                self.context, proc_layer, vads, self.syscall_finder.rule_str
            ):
                syscall_info = self._is_valid_syscall(
                    self.syscall_finder, proc_layer, architecture, vads, address
                )
                if not syscall_info:
                    continue

                address, disasm_bytes = syscall_info

                yield (
                    0,
                    (
                        proc_name,
                        proc.UniqueProcessId,
                        vad_path,
                        format_hints.Hex(address),
                        disasm_bytes,
                    ),
                )

    def run(self) -> renderers.TreeGrid:
        return renderers.TreeGrid(
            [
                ("Process", str),
                ("PID", int),
                ("Range", str),
                ("Address", format_hints.Hex),
                ("Disasm", str),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import Iterable

from volatility3.framework import interfaces
from volatility3.framework.configuration import requirements
from volatility3.plugins.windows import poolscanner, modules, pedump

vollog = logging.getLogger(__name__)


class ModScan(modules.Modules):
    """Scans for modules present in a particular windows memory image."""

    _required_framework_version = (2, 0, 0)

    # 3.0.0 changed the signature of enumeration methods (scan_modules)
    _version = (3, 0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._enumeration_method = self.scan_modules

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="poolscanner", component=poolscanner.PoolScanner, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="modules", component=modules.Modules, version=(3, 0, 0)
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed modules",
                default=False,
                optional=True,
            ),
            requirements.IntRequirement(
                name="base",
                description="Extract a single module with BASE address",
                optional=True,
            ),
            requirements.StringRequirement(
                name="name",
                description="module name/sub string",
                optional=True,
                default=None,
            ),
            requirements.VersionRequirement(
                name="pedump", component=pedump.PEDump, version=(2, 0, 0)
            ),
        ]

    @classmethod
    def scan_modules(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Scans for modules using the poolscanner module and constraints.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: Name of the module for the kernel
        Returns:
            A list of kernel module objects as found from the primary (kernel) layer based on module pool signatures
        """

        kernel = context.modules[kernel_module_name]

        constraints = poolscanner.PoolScanner.builtin_constraints(
            kernel.symbol_table_name, [b"MmLd"]
        )

        for result in poolscanner.PoolScanner.generate_pool_scan(
            context, kernel_module_name, constraints
        ):
            _constraint, mem_object, _header = result
            yield mem_object



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""A module containing a plugin for enumerating memory-mapped
ELF files across all processes."""

import logging
from typing import List, Optional, Type

from volatility3.framework import constants, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.linux.extensions import elf
from volatility3.framework.constants import linux as linux_constants
from volatility3.plugins.linux import pslist


vollog = logging.getLogger(__name__)


class Elfs(plugins.PluginInterface):
    """Lists all memory mapped ELF files for all processes."""

    _required_framework_version = (2, 0, 0)
    _version = (2, 0, 3)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed processes",
                default=False,
                optional=True,
            ),
        ]

    @classmethod
    def elf_dump(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        elf_table_name: str,
        vma: interfaces.objects.ObjectInterface,
        task: interfaces.objects.ObjectInterface,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
    ) -> Optional[interfaces.plugins.FileHandlerInterface]:
        """Extracts an ELF as a FileHandlerInterface
        Args:
            context: the context to operate upon
            layer_name: The name of the layer on which to operate
            elf_table_name: the name for the symbol table containing the symbols for ELF-files
            vma: virtual memory allocation of ELF
            task: the task object whose memory should be output
            open_method: class to provide context manager for opening the file
        Returns:
            An open FileHandlerInterface object containing the complete data for the task or None in the case of failure
        """

        proc_layer = context.layers[layer_name]
        file_handle = None

        elf_object = context.object(
            elf_table_name + constants.BANG + "Elf",
            offset=vma.vm_start,
            layer_name=layer_name,
        )

        if not elf_object.is_valid():
            vollog.debug("ELF object to be dumped is not valid")
            return None

        sections = {}
        # TODO: Apply more effort to reconstruct ELF, e.g.: https://github.com/enbarberis/core2ELF64 ?
        for phdr in elf_object.get_program_headers():
            try:
                if phdr.p_type.description != "PT_LOAD":
                    continue
            except ValueError:
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"Skipping unknown ELF program header type: {phdr.p_type}",
                )
                continue

            start = phdr.p_vaddr
            size = phdr.p_memsz
            end = start + size

            # Use complete memory pages for dumping
            # If start isn't a multiple of a page, stick to the highest multiple < start
            # If end isn't a multiple of a page, stick to the lowest multiple > end
            if start % proc_layer.page_size:
                start = start & proc_layer.page_mask

            if end % proc_layer.page_size:
                end = (end & proc_layer.page_mask) + proc_layer.page_size

            real_size = end - start

            # Check if ELF has a legitimate size
            if real_size < 0 or real_size > linux_constants.ELF_MAX_EXTRACTION_SIZE:
                raise ValueError(f"The claimed size of the ELF is invalid: {real_size}")

            sections[start] = real_size

        elf_data = b""
        for section_start in sorted(sections.keys()):
            read_size = sections[section_start]

            buf = proc_layer.read(vma.vm_start + section_start, read_size, pad=True)
            elf_data = elf_data + buf

        file_handle = open_method(
            f"pid.{task.pid}.{utility.array_to_string(task.comm)}.{vma.vm_start:#x}.dmp"
        )
        file_handle.write(elf_data)

        return file_handle

    def _generator(self, tasks):
        elf_table_name = intermed.IntermediateSymbolTable.create(
            self.context, self.config_path, "linux", "elf", class_types=elf.class_types
        )
        for task in tasks:
            proc_layer_name = task.add_process_layer()
            if not proc_layer_name:
                continue

            proc_layer = self.context.layers[proc_layer_name]

            name = utility.array_to_string(task.comm)

            for vma in task.mm.get_vma_iter():
                hdr = proc_layer.read(vma.vm_start, 4, pad=True)
                if hdr != b"\x7fELF":
                    continue

                path = vma.get_name(self.context, task)

                file_output = "Disabled"
                if self.config["dump"]:
                    file_handle = self.elf_dump(
                        self.context,
                        proc_layer_name,
                        elf_table_name,
                        vma,
                        task,
                        self.open,
                    )
                    file_output = "Error outputting file"
                    if file_handle:
                        file_handle.close()
                        file_output = str(file_handle.preferred_filename)

                yield (
                    0,
                    (
                        task.pid,
                        name,
                        format_hints.Hex(vma.vm_start),
                        format_hints.Hex(vma.vm_end),
                        path or renderers.NotAvailableValue(),
                        file_output,
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Start", format_hints.Hex),
                ("End", format_hints.Hex),
                ("File Path", str),
                ("File Output", str),
            ],
            self._generator(
                pslist.PsList.list_tasks(
                    self.context, self.config["kernel"], filter_func=filter_func
                )
            ),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import hashlib
import json
import logging
import os
import re
from typing import Any, Dict, Optional, Set, Tuple
from volatility3.framework import constants

vollog = logging.getLogger(__name__)

cached_validation_filepath = os.path.join(constants.CACHE_PATH, "valid_isf.hashcache")

validators = {}


def load_cached_validations() -> Set[str]:
    """Loads up the list of successfully cached json objects, so we don't need
    to revalidate them."""
    validhashes: Set = set()
    if os.path.exists(cached_validation_filepath):
        with open(cached_validation_filepath) as f:
            validhashes.update(json.load(f))
    return validhashes


def record_cached_validations(validations: Set[str]) -> None:
    """Record the cached validations, so we don't need to revalidate them in
    future."""
    with open(cached_validation_filepath, "w") as f:
        json.dump(list(validations), f)


cached_validations = load_cached_validations()


def validate(input: Dict[str, Any], use_cache: bool = True) -> bool:
    """Validates an input JSON file based upon."""
    format = input.get("metadata", {}).get("format", None)
    if not format:
        vollog.debug("No schema format defined")
        return False
    basepath = os.path.abspath(os.path.dirname(__file__))
    schema_path = os.path.join(basepath, "schema-" + format + ".json")
    if not os.path.exists(schema_path):
        vollog.debug(f"Schema for format not found: {schema_path}")
        return False
    with open(schema_path) as s:
        schema = json.load(s)
    return valid(input, schema, use_cache)


def create_json_hash(
    input: Dict[str, Any], schema: Optional[Dict[str, Any]] = None
) -> Optional[str]:
    """Constructs the hash of the input and schema to create a unique
    identifier for a particular JSON file."""
    if schema is None:
        format = input.get("metadata", {}).get("format", None)
        if not format:
            vollog.debug("No schema format defined")
            return None
        basepath = os.path.abspath(os.path.dirname(__file__))
        schema_path = os.path.join(basepath, "schema-" + format + ".json")
        if not os.path.exists(schema_path):
            vollog.debug(f"Schema for format not found: {schema_path}")
            return None
        with open(schema_path) as s:
            schema = json.load(s)
    return hashlib.sha1(
        bytes(json.dumps((input, schema), sort_keys=True), "utf-8")
    ).hexdigest()


def valid(
    input: Dict[str, Any], schema: Dict[str, Any], use_cache: bool = True
) -> bool:
    """Validates a json schema."""
    producer = input.get("metadata", {}).get("producer", {})
    if producer and producer.get("name") == "dwarf2json":
        dwarf2json_version = parse_producer_version(producer.get("version", ""))
        # No warnings if version couldn't be parsed, as it's not our role here
        # to validate the schema.
        if dwarf2json_version:
            if dwarf2json_check_rust_type_confusion(input, dwarf2json_version):
                vollog.warning(
                    "This ISF was generated by dwarf2json < 0.9.0, which is known to produce inaccurate results (see dwarf2json GitHub issue #63)."
                )

    input_hash = create_json_hash(input, schema)
    if input_hash in cached_validations and use_cache:
        return True
    try:
        import jsonschema

        schema_key = json.dumps(schema, sort_keys=True)
        if schema_key not in validators:
            validator_class = jsonschema.validators.validator_for(schema)
            validator_class.check_schema(schema)
            validator = validator_class(schema)
            validators[schema_key] = validator
    except ImportError:
        vollog.info("Dependency for validation unavailable: jsonschema")
        vollog.debug("All validations will report success, even with malformed input")
        return True

    try:
        vollog.debug("Validating JSON against schema...")
        validators[schema_key].validate(input)
        cached_validations.add(input_hash)
        vollog.debug("JSON validated against schema (result cached)")
    except jsonschema.exceptions.SchemaError:
        vollog.debug("Schema validation error", exc_info=True)
        return False

    record_cached_validations(cached_validations)
    return True


def parse_producer_version(version_string: str) -> Optional[Tuple[int]]:
    """Parses a producer version and returns a tuple of identifiers.

    Args:
        version_string: string containing dot-separated integers,
    expected to follow the Volatility3 versioning schema

    Returns:
        A tuple containing each version identifier
    """
    identifiers = re.search("^(\\d+)[.](\\d+)[.](\\d+)$", version_string)
    if not identifiers:
        return None

    return tuple(int(d) for d in identifiers.groups())


# dwarf2json sanity checks #
def dwarf2json_check_rust_type_confusion(
    input: Dict[str, Any], dwarf2json_version: Tuple[int]
) -> bool:
    """dwarf2json sanity check for Rust and C types confusion:
     - dwarf2json #63
     - volatility3 #1305

    Args:
        dwarf2json_version: a tuple containing each version identifier

    Returns:
        True if the issue was detected
    """

    return "rust_helper_BUG" in input.get("symbols", {}) and dwarf2json_version < (
        0,
        9,
        0,
    )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0

# Full details on the techniques used in these plugins to detect EDR-evading malware
# can be found in our 20 page whitepaper submitted to DEFCON along with the presentation
# https://www.volexity.com/wp-content/uploads/2024/08/Defcon24_EDR_Evasion_Detection_White-Paper_Andrew-Case.pdf

import logging

from typing import Dict, Tuple, List, Generator

from volatility3.framework import interfaces, exceptions
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.plugins.windows import pslist, pe_symbols

vollog = logging.getLogger(__name__)


class UnhookedSystemCalls(interfaces.plugins.PluginInterface):
    """Detects hooked ntdll.dll stub functions in Windows processes."""

    _required_framework_version = (2, 4, 0)
    _version = (2, 0, 0)

    system_calls = {
        "ntdll.dll": {
            pe_symbols.wanted_names_identifier: [
                "NtCreateThread",
                "NtProtectVirtualMemory",
                "NtReadVirtualMemory",
                "NtOpenProcess",
                "NtWriteFile",
                "NtQueryVirtualMemory",
                "NtAllocateVirtualMemory",
                "NtWorkerFactoryWorkerReady",
                "NtAcceptConnectPort",
                "NtAddDriverEntry",
                "NtAdjustPrivilegesToken",
                "NtAlpcCreatePort",
                "NtClose",
                "NtCreateFile",
                "NtCreateMutant",
                "NtOpenFile",
                "NtOpenIoCompletion",
                "NtOpenJobObject",
                "NtOpenKey",
                "NtOpenKeyEx",
                "NtOpenThread",
                "NtOpenThreadToken",
                "NtOpenThreadTokenEx",
                "NtWriteVirtualMemory",
                "NtTraceEvent",
                "NtTranslateFilePath",
                "NtUmsThreadYield",
                "NtUnloadDriver",
                "NtUnloadKey",
                "NtUnloadKey2",
                "NtUnloadKeyEx",
                "NtCreateKey",
                "NtCreateSection",
                "NtDeleteKey",
                "NtDeleteValueKey",
                "NtDuplicateObject",
                "NtQueryValueKey",
                "NtReplaceKey",
                "NtRequestWaitReplyPort",
                "NtRestoreKey",
                "NtSetContextThread",
                "NtSetSecurityObject",
                "NtSetValueKey",
                "NtSystemDebugControl",
                "NtTerminateProcess",
            ]
        }
    }

    # This data structure is used to track unique implementations of functions across processes
    # The outer dictionary holds the module name (e.g., ntdll.dll)
    # The next dictionary holds the function names (NtTerminateProcess, NtSetValueKey, etc.) inside a module
    # The innermost dictionary holds the unique implementation (bytes) of a function across processes
    # Each implementation is tracked along with the process(es) that host it
    # For systems without malware, all functions should have the same implementation
    # When API hooking/module unhooking is done, the victim (infected) processes will have unique implementations
    _code_bytes_type = Dict[str, Dict[str, Dict[bytes, List[Tuple[int, str]]]]]

    @classmethod
    def get_requirements(cls) -> List:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pe_symbols", component=pe_symbols.PESymbols, version=(3, 0, 0)
            ),
        ]

    def _gather_code_bytes(
        self,
        kernel_module_name: str,
        found_symbols: pe_symbols.found_symbols_type,
    ) -> _code_bytes_type:
        """
        Enumerates the desired DLLs and function implementations in each process
        Groups based on unique implementations of each DLLs' functions
        The purpose is to detect when a function has different implementations (code)
        in different processes.
        This very effectively detects code injection.
        """
        code_bytes: UnhookedSystemCalls._code_bytes_type = {}

        procs = pslist.PsList.list_processes(self.context, kernel_module_name)

        for proc in procs:
            try:
                proc_id = proc.UniqueProcessId
                proc_name = utility.array_to_string(proc.ImageFileName)
                proc_layer_name = proc.add_process_layer()
            except exceptions.InvalidAddressException:
                continue

            for dll_name, functions in found_symbols.items():
                for func_name, func_addr in functions:
                    try:
                        fbytes = self.context.layers[proc_layer_name].read(
                            func_addr, 0x20
                        )
                    except exceptions.InvalidAddressException:
                        continue

                    # see the definition of _code_bytes_type for details of this data structure
                    if dll_name not in code_bytes:
                        code_bytes[dll_name] = {}

                    if func_name not in code_bytes[dll_name]:
                        code_bytes[dll_name][func_name] = {}

                    if fbytes not in code_bytes[dll_name][func_name]:
                        code_bytes[dll_name][func_name][fbytes] = []

                    code_bytes[dll_name][func_name][fbytes].append((proc_id, proc_name))

        return code_bytes

    def _generator(self) -> Generator[Tuple[int, Tuple[str, str, int]], None, None]:
        found_symbols = pe_symbols.PESymbols.addresses_for_process_symbols(
            context=self.context,
            config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
            symbols=UnhookedSystemCalls.system_calls,
        )

        # code_bytes[dll_name][func_name][func_bytes]
        code_bytes = self._gather_code_bytes(self.config["kernel"], found_symbols)

        # walk the functions that were evaluated
        for functions in code_bytes.values():
            # cbb is the distinct groups of bytes (instructions)
            # for this function across processes
            for func_name, cbb in functions.items():
                # the dict key here is the raw instructions, which is not helpful to look at
                # the values are the list of tuples for the (proc_id, proc_name) pairs for this set of bytes (instructions)
                cb = list(cbb.values())

                # if all processes map to the same implementation, then no malware is present
                if len(cb) == 1:
                    yield 0, (func_name, "", len(cb[0]))
                else:
                    # if there are differing implementations then it means
                    # that malware has overwritten system call(s) in infected processes
                    # max_idx and small_idx find which implementation of a system call has the least processes
                    # as all observed malware and open source projects only infected a few targets, leaving the
                    # rest with the original EDR hooks in place
                    max_idx = 0 if len(cb[0]) > len(cb[1]) else 1
                    small_idx = (~max_idx) & 1

                    ps = []

                    # gather processes on small_idx since these are the malware infected ones
                    for pid, pname in cb[small_idx]:
                        ps.append(f"{pid:d}:{pname}")

                    proc_names = ", ".join(ps)

                    yield 0, (func_name, proc_names, len(cb[max_idx]))

    def run(self) -> renderers.TreeGrid:
        return renderers.TreeGrid(
            [
                ("Function", str),
                ("Distinct Implementations", str),
                ("Total Implementations", int),
            ],
            self._generator(),
        )



##
## plugin for testing addition of threads scan support to poolscanner.py
##
import datetime
import logging
from typing import Callable, Dict, NamedTuple, Optional, Union, Tuple, Iterator

from volatility3.framework import exceptions, interfaces, objects, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.constants import windows as windows_constants
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols.windows import extensions as win_extensions
from volatility3.plugins import timeliner
from volatility3.plugins.windows import pe_symbols, poolscanner

vollog = logging.getLogger(__name__)


class ThrdScan(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Scans for windows threads."""

    # version 2.6.0 adds support for scanning for 'Ethread' structures by pool tags
    _required_framework_version = (2, 6, 0)
    _version = (2, 1, 0)

    class ThreadInfo(NamedTuple):
        offset: int
        pid: objects.Pointer
        tid: objects.Pointer
        start_addr: objects.Pointer
        start_path: Optional[str]
        win32_start_addr: objects.Pointer
        win32_start_path: Optional[str]
        create_time: Union[datetime.datetime, interfaces.renderers.BaseAbsentValue]
        exit_time: Union[datetime.datetime, interfaces.renderers.BaseAbsentValue]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.implementation = self.scan_threads

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="poolscanner", component=poolscanner.PoolScanner, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pe_symbols", component=pe_symbols.PESymbols, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
        ]

    @classmethod
    def scan_threads(
        cls,
        context: interfaces.context.ContextInterface,
        module_name: str,
    ) -> Iterator[win_extensions.ETHREAD]:
        """Scans for threads using the poolscanner module and constraints.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            module_name: Name of the module to use for scanning

        Returns:
              A list of _ETHREAD objects found by scanning memory for the "Thre" / "Thr\\xE5" pool signatures
        """

        kernel = context.modules[module_name]

        constraints = poolscanner.PoolScanner.builtin_constraints(
            kernel.symbol_table_name, [b"Thr\xe5", b"Thre"]
        )

        for result in poolscanner.PoolScanner.generate_pool_scan(
            context, module_name, constraints
        ):
            _constraint, mem_object, _header = result
            yield mem_object

    @classmethod
    def gather_thread_info(
        cls,
        ethread: win_extensions.ETHREAD,
        vads_cache: Optional[Dict[int, pe_symbols.ranges_type]] = None,
    ) -> Optional[ThreadInfo]:
        try:
            thread_offset = ethread.vol.offset
            owner_proc_pid = ethread.Cid.UniqueProcess
            thread_tid = ethread.Cid.UniqueThread
            thread_start_addr = ethread.StartAddress
            thread_win32start_addr = ethread.Win32StartAddress
            thread_create_time = ethread.get_create_time()  # datetime.datetime object / volatility3.framework.renderers.UnparsableValue object
            thread_exit_time = ethread.get_exit_time()  # datetime.datetime object / volatility3.framework.renderers.UnparsableValue object

            owner_proc = None
            if vads_cache is not None:
                owner_proc = ethread.owning_process()
        except exceptions.InvalidAddressException:
            vollog.debug(f"Thread invalid address {ethread.vol.offset:#x}")
            return None

        # Filter junk PIDs
        if (
            ethread.Cid.UniqueProcess > windows_constants.MAX_PID
            or ethread.Cid.UniqueProcess == 0
            or ethread.Cid.UniqueProcess % 4 != 0
        ):
            return None

        # Get VAD mappings for valid non-system (PID 4) processes
        if (
            owner_proc
            and owner_proc.is_valid()
            and owner_proc.UniqueProcessId != 4
            and vads_cache is not None
        ):
            vads = pe_symbols.PESymbols.get_vads_for_process_cache(
                vads_cache, owner_proc
            )

            start_path = (
                pe_symbols.PESymbols.filepath_for_address(vads, thread_start_addr)
                if vads
                else None
            )
            win32start_path = (
                pe_symbols.PESymbols.filepath_for_address(vads, thread_win32start_addr)
                if vads
                else None
            )
        else:
            start_path = None
            win32start_path = None

        return cls.ThreadInfo(
            thread_offset,
            owner_proc_pid,
            thread_tid,
            thread_start_addr,
            start_path,
            thread_win32start_addr,
            win32start_path,
            thread_create_time,
            thread_exit_time,
        )

    def _generator(self, filter_func: Callable) -> Iterator[Tuple[int, Tuple]]:
        kernel_name = self.config["kernel"]

        vads_cache: Dict[int, pe_symbols.ranges_type] = {}

        for ethread in self.implementation(self.context, kernel_name):
            info = self.gather_thread_info(ethread, vads_cache)

            if info:
                yield (
                    0,
                    (
                        format_hints.Hex(info.offset),
                        info.pid,
                        info.tid,
                        format_hints.Hex(info.start_addr),
                        info.start_path or renderers.NotAvailableValue(),
                        format_hints.Hex(info.win32_start_addr),
                        info.win32_start_path or renderers.NotAvailableValue(),
                        info.create_time,
                        info.exit_time,
                    ),
                )

    def generate_timeline(self):
        filt_func = self.filter_func(self.config)

        for row in self._generator(filt_func):
            _depth, row_data = row
            row_dict = {}
            (
                row_dict["Offset"],
                row_dict["PID"],
                row_dict["TID"],
                row_dict["StartAddress"],
                row_dict["StartPath"],
                row_dict["Win32StartAddress"],
                row_dict["Win32StartPath"],
                row_dict["CreateTime"],
                row_dict["ExitTime"],
            ) = row_data

            # Skip threads with no creation time
            # - mainly system process threads
            if not isinstance(row_dict["CreateTime"], datetime.datetime):
                continue
            description = f"Thread: Tid {row_dict['TID']} in Pid {row_dict['PID']} (Offset {row_dict['Offset']})"

            # yield created time, and if there is exit time, yield it too.
            yield (description, timeliner.TimeLinerType.CREATED, row_dict["CreateTime"])
            if isinstance(row_dict["ExitTime"], datetime.datetime):
                yield (
                    description,
                    timeliner.TimeLinerType.MODIFIED,
                    row_dict["ExitTime"],
                )

    @classmethod
    def filter_func(cls, config: interfaces.configuration.HierarchicalDict) -> Callable:
        """Returns a function that can filter this plugin's implementation method based on the config"""
        return lambda x: False

    def run(self):
        filt_func = self.filter_func(self.config)

        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("PID", int),
                ("TID", int),
                ("StartAddress", format_hints.Hex),
                ("StartPath", str),
                ("Win32StartAddress", format_hints.Hex),
                ("Win32StartPath", str),
                ("CreateTime", datetime.datetime),
                ("ExitTime", datetime.datetime),
            ],
            self._generator(filt_func),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import re
from typing import List

from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.layers import scanners
from volatility3.framework.renderers import format_hints

vollog = logging.getLogger(__name__)


class RegExScan(plugins.PluginInterface):
    """Scans kernel memory using RegEx patterns."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)
    MAXSIZE_DEFAULT = 128

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.TranslationLayerRequirement(
                name="primary",
                description="Memory layer for the kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.StringRequirement(
                name="pattern", description="RegEx pattern", optional=False
            ),
            requirements.IntRequirement(
                name="maxsize",
                description="Maximum size in bytes for displayed context",
                default=cls.MAXSIZE_DEFAULT,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="regex_scanner",
                component=scanners.RegExScanner,
                version=(1, 0, 0),
            ),
        ]

    def _generator(self, context, layer_name, pattern, maxsize):
        layer = self.context.layers[layer_name]
        vollog.debug(f"RegEx Pattern: {pattern}")

        # Convert string pattern to bytes for RegExScanner
        pattern_bytes = pattern.encode("utf-8")

        # Compile the pattern here to ensure consistency
        try:
            compiled_pattern = re.compile(pattern_bytes)
        except re.error as e:
            vollog.error(f"Invalid regex pattern: {e}")
            raise ValueError(f"Invalid regex pattern: {e}")

        for offset in layer.scan(
            context=context, scanner=scanners.RegExScanner(pattern_bytes)
        ):
            result_data = layer.read(offset, maxsize, pad=True)

            # reapply the regex in order to extract just the match
            regex_result = compiled_pattern.search(result_data)

            if regex_result:
                # the match is within the results_data (e.g. it fits within maxsize)
                # extract just the match itself
                regex_match = regex_result.group(0)
                text_result = str(regex_match, encoding="UTF-8", errors="replace")
                bytes_result = regex_match
            else:
                # the match is not with the results_data (e.g. it doesn't fit within maxsize)
                text_result = str(result_data, encoding="UTF-8", errors="replace")
                bytes_result = result_data

            yield 0, (format_hints.Hex(offset), text_result, bytes_result)

    def run(self):
        pattern = self.config.get("pattern")
        maxsize = self.config.get("maxsize", self.MAXSIZE_DEFAULT)
        layer_name = self.config["primary"]
        context = self.context

        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Text", str),
                ("Hex", bytes),
            ],
            self._generator(context, layer_name, pattern, maxsize),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""A module containing a plugin that verifies the operation function
pointers of network protocols."""

import logging
from typing import List, Tuple, Generator

from volatility3.framework import exceptions, interfaces
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.renderers import format_hints

vollog = logging.getLogger(__name__)


class Check_afinfo(plugins.PluginInterface):
    """Verifies the operation function pointers of network protocols."""

    _version = (1, 0, 0)
    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
        ]

    @classmethod
    def _check_members(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_name: str,
        var_ops: interfaces.objects.ObjectInterface,
        var_name: str,
        members: List[str],
    ) -> Generator[Tuple[str, str, int], None, None]:
        """
        Yields any members that are not pointing inside the kernel
        """

        vmlinux = context.modules[vmlinux_name]

        for check in members:
            # redhat-specific garbage
            if check.startswith("__UNIQUE_ID_rh_kabi_hide"):
                continue

            # These structures have members like `write` and `next`, which are built in Python functions
            addr = var_ops.member(attr=check)

            # Unimplemented handlers are set to 0
            if not addr:
                continue

            if len(vmlinux.get_symbols_by_absolute_location(addr)) == 0:
                yield var_name, check, addr

    @classmethod
    def _check_pre_4_18_ops(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_name: str,
        var_name: str,
        var: interfaces.objects.ObjectInterface,
        op_members: List[str],
        seq_members: List[str],
    ):
        """
        Finds the correct way to reference `op_members`
        """
        vmlinux = context.modules[vmlinux_name]

        if var.has_member("seq_fops"):
            yield from cls._check_members(
                context, vmlinux_name, var.seq_fops, var_name, op_members
            )
        # newer kernels
        if var.has_member("seq_ops"):
            yield from cls._check_members(
                context, vmlinux_name, var.seq_ops, var_name, seq_members
            )

        # this is the most commonly hooked member by rootkits, so a force a check on it
        elif var.has_member("seq_show"):
            if len(vmlinux.get_symbols_by_location(var.seq_show)) == 0:
                yield var_name, "show", var.seq_show
        else:
            raise exceptions.VolatilityException(
                "_check_afinfo_pre_4_18: Unable to find sequence operations members for checking."
            )

    @classmethod
    def _check_afinfo_pre_4_18(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_name: str,
        seq_members: str,
    ) -> Generator[Tuple[str, str, int], None, None]:
        """
        Checks the operations structures for network protocols of < 4.18 systems
        """
        tcp = ("tcp_seq_afinfo", ["tcp6_seq_afinfo", "tcp4_seq_afinfo"])
        udp = (
            "udp_seq_afinfo",
            [
                "udplite6_seq_afinfo",
                "udp6_seq_afinfo",
                "udplite4_seq_afinfo",
                "udp4_seq_afinfo",
            ],
        )
        protocols = [tcp, udp]

        vmlinux = context.modules[vmlinux_name]

        op_members = vmlinux.get_type("file_operations").members

        # loop through all symbols
        for struct_type, global_vars in protocols:
            for global_var_name in global_vars:
                # this will lookup fail for the IPv6 protocols on kernels without IPv6 support
                try:
                    global_var = vmlinux.object_from_symbol(global_var_name)
                except exceptions.SymbolError:
                    continue

                yield from cls._check_pre_4_18_ops(
                    context,
                    vmlinux_name,
                    global_var_name,
                    global_var,
                    op_members,
                    seq_members,
                )

    @classmethod
    def _check_afinfo_post_4_18(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_name: str,
        seq_members: str,
    ) -> Generator[Tuple[str, str, int], None, None]:
        """
        Checks the operations structures for network protocols of >= 4.18 systems
        """
        vmlinux = context.modules[vmlinux_name]

        ops_structs = [
            "raw_seq_ops",
            "udp_seq_ops",
            "arp_seq_ops",
            "unix_seq_ops",
            "udp6_seq_ops",
            "raw6_seq_ops",
            "tcp_seq_ops",
            "tcp4_seq_ops",
            "tcp6_seq_ops",
            "packet_seq_ops",
        ]

        for protocol_ops_var in ops_structs:
            # These will fail if the particular kernel doesn't have support for a protocol like IPv6
            try:
                protocol_ops = vmlinux.object_from_symbol(protocol_ops_var)
            except exceptions.SymbolError:
                continue

            yield from cls._check_members(
                context, vmlinux_name, protocol_ops, protocol_ops_var, seq_members
            )

    @classmethod
    def check_afinfo(
        cls, context: interfaces.context.ContextInterface, vmlinux_name
    ) -> Generator[Tuple[str, str, int], None, None]:
        """
        Walks the network protocol operations structures for common network protocols.
        Reports any initialized operations members that do not point inside the kernel.
        """
        vmlinux = context.modules[vmlinux_name]

        type_check = vmlinux.get_type("tcp_seq_afinfo")
        if type_check.has_member("seq_fops"):
            checker = cls._check_afinfo_pre_4_18
        else:
            checker = cls._check_afinfo_post_4_18

        seq_members = vmlinux.get_type("seq_operations").members

        yield from checker(context, vmlinux_name, seq_members)

    def _generator(self):
        """
        A simple wrapper around `check_afino`
        """
        for name, member, address in self.check_afinfo(
            self.context, self.config["kernel"]
        ):
            yield 0, (name, member, format_hints.Hex(address))

    def run(self):
        return renderers.TreeGrid(
            [
                ("Symbol Name", str),
                ("Member", str),
                ("Handler Address", format_hints.Hex),
            ],
            self._generator(),
        )



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0

import json
import logging
import ntpath
import os
import re
from typing import List, Dict, Union

from volatility3.framework import (
    renderers,
    interfaces,
    objects,
    exceptions,
    constants,
    layers,
)
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols.windows.extensions import registry
from volatility3.plugins.windows import pslist
from volatility3.plugins.windows.registry import hivelist

vollog = logging.getLogger(__name__)


def find_sid_re(
    sid_string, sid_re_list
) -> Union[str, interfaces.renderers.BaseAbsentValue]:
    for reg, name in sid_re_list:
        if reg.search(sid_string):
            return name
    return renderers.NotAvailableValue()


class GetSIDs(interfaces.plugins.PluginInterface):
    """Print the SIDs owning each process"""

    _version = (1, 0, 0)
    _required_framework_version = (2, 0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for plugin_dir in constants.PLUGINS_PATH:
            sids_json_file_name = os.path.join(
                plugin_dir, os.path.join("windows", "sids_and_privileges.json")
            )
            if os.path.exists(sids_json_file_name):
                break
        else:
            vollog.log(
                constants.LOGLEVEL_VVV,
                "sids_and_privileges.json file is missing plugin error",
            )
            raise RuntimeError(
                "The sids_and_privileges.json file missed from you plugin directory"
            )

        # Get all the sids from the json file.
        with open(sids_json_file_name) as file_handle:
            sids_json_data = json.load(file_handle)
            self.servicesids = sids_json_data["service sids"]
            self.well_known_sids = sids_json_data["well known"]

            # Compile all the sids regex.
            self.well_known_sid_re = [
                (re.compile(c_list[0]), c_list[1])
                for c_list in sids_json_data["sids re"]
            ]

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
        ]

    def lookup_user_sids(self) -> Dict[str, str]:
        """
        Enumerate the registry for all the users.

        Returns:
            An dictionary of {sid: user name}
        """

        key = "Microsoft\\Windows NT\\CurrentVersion\\ProfileList"
        val = "ProfileImagePath"

        sids = {}
        for hive in hivelist.HiveList.list_hives(
            context=self.context,
            base_config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
            filter_string="config\\software",
            hive_offsets=None,
        ):
            try:
                for subkey in hive.get_key(key).get_subkeys():
                    try:
                        sid = str(subkey.get_name())
                    except (
                        exceptions.InvalidAddressException,
                        layers.registry.RegistryException,
                    ):
                        continue

                    path = ""
                    for node in subkey.get_values():
                        try:
                            value_node_name = node.get_name() or "(Default)"
                        except (
                            exceptions.InvalidAddressException,
                            layers.registry.RegistryException,
                        ):
                            continue
                        try:
                            value_data = node.decode_data()
                            if isinstance(value_data, int):
                                value_data = format_hints.MultiTypeData(
                                    value_data, encoding="utf-8"
                                )
                            elif (
                                registry.RegValueTypes(node.Type)
                                == registry.RegValueTypes.REG_BINARY
                            ):
                                value_data = format_hints.MultiTypeData(
                                    value_data, show_hex=True
                                )
                            elif (
                                registry.RegValueTypes(node.Type)
                                == registry.RegValueTypes.REG_MULTI_SZ
                            ):
                                value_data = format_hints.MultiTypeData(
                                    value_data, encoding="utf-16-le", split_nulls=True
                                )
                            else:
                                value_data = format_hints.MultiTypeData(
                                    value_data, encoding="utf-16-le"
                                )
                            if value_node_name == val:
                                path = str(value_data).replace("\\x00", "")[:-1]
                                user = ntpath.basename(path)
                                sids[sid] = user
                        except (
                            ValueError,
                            exceptions.InvalidAddressException,
                            layers.registry.RegistryException,
                        ):
                            continue
            except (
                KeyError,
                exceptions.InvalidAddressException,
                layers.registry.RegistryException,
            ):
                continue

        return sids

    def _generator(self, procs):
        user_sids = self.lookup_user_sids()

        # Go all over the process list, get the token
        for task in procs:
            # Make sure we have a valid token
            try:
                token = task.Token.dereference().cast("_TOKEN")
            except exceptions.InvalidAddressException:
                token = False

            task_name = objects.utility.array_to_string(task.ImageFileName)

            if not token or not isinstance(token, interfaces.objects.ObjectInterface):
                yield (
                    0,
                    [
                        int(task.UniqueProcessId),
                        task_name,
                        "Token unreadable",
                        "",
                    ],
                )
                continue

            # Go all over the sids and try to translate them with one of the tables we have
            for sid_string in token.get_sids():
                if sid_string in self.well_known_sids:
                    sid_name = self.well_known_sids[sid_string]
                elif sid_string in self.servicesids:
                    sid_name = self.servicesids[sid_string]
                elif sid_string in user_sids:
                    sid_name = user_sids[sid_string]
                else:
                    sid_name_re = find_sid_re(sid_string, self.well_known_sid_re)
                    if sid_name_re:
                        sid_name = sid_name_re
                    else:
                        sid_name = ""

                yield (
                    0,
                    (
                        task.UniqueProcessId,
                        task_name,
                        sid_string,
                        sid_name,
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [("PID", int), ("Process", str), ("SID", str), ("Name", str)],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=filter_func,
                )
            ),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0

import logging
import io
import pefile
from volatility3.framework.symbols import intermed
from volatility3.framework import renderers, interfaces, exceptions, constants
from volatility3.framework.configuration import requirements
from volatility3.plugins.windows import pslist
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols.windows.extensions import pe

vollog = logging.getLogger(__name__)


class IAT(interfaces.plugins.PluginInterface):
    """Extract Import Address Table to list API (functions) used by a program contained in external libraries"""

    _required_framework_version = (2, 4, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process ID to include (all other processes are excluded)",
                optional=True,
            ),
        ]

    def _generator(self, procs):
        kernel = self.context.modules[self.config["kernel"]]

        for proc in procs:
            try:
                proc_id = proc.UniqueProcessId
                proc_layer_name = proc.add_process_layer()
                peb = self.context.object(
                    kernel.symbol_table_name + constants.BANG + "_PEB",
                    layer_name=proc_layer_name,
                    offset=proc.Peb,
                )

                if proc_layer_name is None:
                    raise TypeError("add_process_layer failed")

                pe_table_name = intermed.IntermediateSymbolTable.create(
                    self.context,
                    self.config_path,
                    "windows",
                    "pe",
                    class_types=pe.class_types,
                )
                pe_data = io.BytesIO()

                dos_header = self.context.object(
                    pe_table_name + constants.BANG + "_IMAGE_DOS_HEADER",
                    offset=peb.ImageBaseAddress,
                    layer_name=proc_layer_name,
                )

                try:
                    for offset, data in dos_header.reconstruct():
                        pe_data.seek(offset)
                        pe_data.write(data)
                except (exceptions.InvalidAddressException, ValueError) as excp:
                    vollog.warning(
                        f"Exception triggered when reconstructing PE file for process {proc.UniqueProcessId} at address {peb.ImageBaseAddress:#x} due to {excp}. Output file may be corrupt and/or truncated."
                    )

                try:
                    pe_obj = pefile.PE(data=pe_data.getvalue(), fast_load=True)
                except pefile.PEFormatError as excp:
                    vollog.debug(
                        f"Exception triggered when creating PE file object for process {proc.UniqueProcessId} at address {peb.ImageBaseAddress:#x} due to {excp}. Unable to extract file."
                    )
                    continue

                pe_obj.parse_data_directories(
                    [pefile.DIRECTORY_ENTRY["IMAGE_DIRECTORY_ENTRY_IMPORT"]]
                )
                if hasattr(pe_obj, "DIRECTORY_ENTRY_IMPORT"):
                    for entry in pe_obj.DIRECTORY_ENTRY_IMPORT:
                        dll_entry = entry.dll
                        if dll_entry:
                            dll_entry = dll_entry.decode()
                        else:
                            dll_entry = renderers.NotAvailableValue

                        bound = True
                        # Initially set to 0 if not bound
                        time_date_stamp = entry.struct.TimeDateStamp
                        if not time_date_stamp:
                            bound = False

                        # Iterate over imported functions
                        for imp in entry.imports:
                            import_name = imp.name
                            if import_name:
                                import_name = imp.name.decode()
                            else:
                                import_name = renderers.NotAvailableValue()
                            function_address = (
                                pe_obj.OPTIONAL_HEADER.ImageBase + imp.address
                            )
                            if not function_address:
                                function_address = renderers.NotAvailableValue

                            yield (
                                0,
                                (
                                    proc_id,
                                    proc.ImageFileName.cast(
                                        "string",
                                        max_length=proc.ImageFileName.vol.count,
                                        errors="replace",
                                    ),
                                    dll_entry,
                                    bound,
                                    import_name,
                                    format_hints.Hex(function_address),
                                ),
                            )
            except exceptions.InvalidAddressException as excp:
                vollog.debug(
                    f"Process {proc_id}: invalid address {excp.invalid_address} in layer {excp.layer_name}"
                )
                continue

    def run(self):
        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Name", str),
                ("Library", str),
                ("Bound", bool),
                ("Function", str),
                ("Address", format_hints.Hex),
            ],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=pslist.PsList.create_pid_filter(
                        self.config.get("pid", None)
                    ),
                )
            ),
        )



import contextlib
import logging
import struct
from typing import Iterator, List, Optional, Tuple, Type

from volatility3.framework import exceptions, interfaces, renderers
from volatility3.framework.layers import registry as registry_layer
from volatility3.framework.configuration import requirements
from volatility3.framework.symbols.windows.extensions import registry
from volatility3.plugins.windows.registry import hivelist, printkey

vollog = logging.getLogger(__name__)


class Certificates(interfaces.plugins.PluginInterface):
    """Lists the certificates in the registry's Certificate Store."""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="printkey", component=printkey.PrintKey, version=(1, 0, 0)
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed certificates",
                default=False,
                optional=True,
            ),
        ]

    def parse_data(self, data: bytes) -> Tuple[str, bytes]:
        name = renderers.NotAvailableValue()
        certificate_data = renderers.NotAvailableValue()
        while len(data) > 12:
            ctype, clength = struct.unpack("<QI", data[0:12])
            cvalue, data = data[12 : 12 + clength], data[12 + clength :]
            if ctype == 0x10000000B:
                name = str(cvalue, "utf-16").strip("\x00")
            elif ctype == 0x100000020:
                certificate_data = cvalue
        return (name, certificate_data)

    @classmethod
    def dump_certificate(
        cls,
        certificate_data: bytes,
        hive_offset: int,
        reg_section: str,
        key_hash: str,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
    ) -> Optional[interfaces.plugins.FileHandlerInterface]:
        try:
            dump_name = f"{hive_offset}-{reg_section}-{key_hash}.crt"
            file_handle = open_method(dump_name)
            file_handle.write(certificate_data)
            return file_handle
        except exceptions.InvalidAddressException:
            vollog.debug(f"Unable to dump certificate file at {hive_offset:#x}")
        return None

    def _generator(self) -> Iterator[Tuple[int, Tuple[str, str, str, str]]]:
        for hive in hivelist.HiveList.list_hives(
            context=self.context,
            base_config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
        ):
            for top_key in [
                "Microsoft\\SystemCertificates",
                "Software\\Microsoft\\SystemCertificates",
            ]:
                with contextlib.suppress(
                    KeyError,
                    registry_layer.RegistryException,
                    exceptions.InvalidAddressException,
                ):
                    # Walk it
                    node_path = hive.get_key(top_key, return_list=True)
                    for (
                        _depth,
                        is_key,
                        _last_write_time,
                        key_path,
                        _volatility,
                        node,
                    ) in printkey.PrintKey.key_iterator(hive, node_path, recurse=True):
                        if (
                            not is_key
                            and registry.RegValueTypes(node.Type)
                            == registry.RegValueTypes.REG_BINARY
                        ):
                            name, certificate_data = self.parse_data(node.decode_data())
                            unique_key_offset = (
                                key_path.casefold().index(top_key.casefold())
                                + len(top_key)
                                + 1
                            )
                            reg_section = key_path[
                                unique_key_offset : key_path.index(
                                    "\\", unique_key_offset
                                )
                            ]
                            key_hash = key_path[key_path.rindex("\\") + 1 :]

                            if self.config["dump"]:
                                if not isinstance(
                                    certificate_data,
                                    interfaces.renderers.BaseAbsentValue,
                                ):
                                    file_handle = self.dump_certificate(
                                        certificate_data,
                                        hive.hive_offset,
                                        reg_section,
                                        key_hash,
                                        self.open,
                                    )
                                    if file_handle:
                                        file_handle.close()

                            yield (0, (top_key, reg_section, key_hash, name))

    def run(self) -> renderers.TreeGrid:
        return renderers.TreeGrid(
            [
                ("Certificate path", str),
                ("Certificate section", str),
                ("Certificate ID", str),
                ("Certificate name", str),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import base64
import codecs
import copy
import json
import logging
import os
import pathlib
import zipfile
from abc import ABCMeta
from typing import Any, Dict, Generator, Iterable, List, Mapping, Optional, Tuple, Type

from volatility3 import schemas, symbols
from volatility3.framework import (
    class_subclasses,
    constants,
    exceptions,
    interfaces,
    objects,
)
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import resources
from volatility3.framework.symbols import metadata, native

vollog = logging.getLogger(__name__)


# ## TODO
#
# All symbol tables should take a label to an object template
#
# Templates for subtypes etc should be looked up recursively just like anything else
# We therefore need a way to unroll rolled-up types
# Generate mangled names on the fly (prohibits external calling)
#
# Symbol list could be a dict with knowledge of its parent?
# Class split is arbitrary, it's an extension for developers
# Object template should contain both class and initial parameters
#
#
# *** Resolution should not happen in the resolve function
# It should only happen on access of contained types ***
#
# Recursive objects can be fixed by having caching the objects
# (however, they have to be built first!)
#
# Single hop resolution is probably the solution
# Could probably deal with it by having a property that caches
# for container types
#


def _construct_delegate_function(name: str, is_property: bool = False) -> Any:
    def _delegate_function(self, *args, **kwargs):
        if is_property:
            return getattr(self._delegate, name)
        return getattr(self._delegate, name)(*args, **kwargs)

    if is_property:
        return property(_delegate_function)
    return _delegate_function


class IntermediateSymbolTable(interfaces.symbols.SymbolTableInterface):
    """The IntermediateSymbolTable class reads a JSON file and conducts common
    tasks such as validation, construction by looking up a JSON file from the
    available files and ensuring the appropriate version of the schema and
    proxy are chosen.

    The JSON format itself is made up of various groups (symbols, user_types, base_types, enums and metadata)
        * Symbols link a name to a particular offset relative to the start of a section of memory
        * Base types define the simplest primitive data types, these can make more complex structure
        * User types define the more complex types by specifying members at a relative offset from the start of the type
        * Enums can specify a list of names and values and a type inside which the numeric encoding will fit
        * Metadata defines information about the originating file

    These are documented in JSONSchema JSON files located in volatility3/schemas.
    """

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        name: str,
        isf_url: str,
        native_types: Optional[interfaces.symbols.NativeTableInterface] = None,
        table_mapping: Optional[Dict[str, str]] = None,
        validate: bool = True,
        class_types: Optional[
            Mapping[str, Type[interfaces.objects.ObjectInterface]]
        ] = None,
        symbol_mask: int = 0,
    ) -> None:
        """Instantiates a SymbolTable based on an IntermediateSymbolFormat JSON file.  This is validated against the
        appropriate schema.  The validation can be disabled by passing validate = False, but this should almost never be
        done.

        Args:
            context: The volatility context for the symbol table
            config_path: The configuration path for the symbol table
            name: The name for the symbol table (this is used in symbols e.g. table!symbol)
            isf_url: The URL pointing to the ISF file location
            native_types: The NativeSymbolTable that contains the native types for this symbol table
            table_mapping: A dictionary linking names referenced in the file with symbol tables in the context
            validate: Determines whether the ISF file will be validated against the appropriate schema
            class_types: A dictionary of type names and classes that override StructType when they are instantiated
            symbol_mask: An address mask used for all returned symbol offsets from this table (a mask of 0 disables masking)
        """
        # Check there are no obvious errors
        # Open the file and test the version
        self._versions = dict((x.version, x) for x in class_subclasses(ISFormatTable))
        with resources.ResourceAccessor().open(isf_url) as fp:
            reader = codecs.getreader("utf-8")
            json_object = json.load(reader(fp))  # type: ignore

        # Validation is expensive, but we cache to store the hashes of successfully validated json objects
        if validate and not schemas.validate(json_object):
            raise exceptions.SymbolSpaceError(
                f"File does not pass version validation: {isf_url}"
            )

        metadata = json_object.get("metadata", None)

        if not metadata:
            raise exceptions.SymbolSpaceError(
                f"Invalid ISF file attempted to be parsed: {isf_url}"
            )

        # Determine the delegate or throw an exception
        self._delegate = self._closest_version(
            metadata.get("format", "0.0.0"), self._versions
        )(context, config_path, name, json_object, native_types, table_mapping)
        if self._delegate.version < constants.ISF_MINIMUM_SUPPORTED:
            raise RuntimeError(
                "ISF version {} is no longer supported: {}".format(
                    metadata.get("format", "0.0.0"), isf_url
                )
            )
        elif self._delegate.version < constants.ISF_MINIMUM_DEPRECATED:
            vollog.warning(
                f"ISF version {metadata.get('format', '0.0.0')} has been deprecated: {isf_url}"
            )

        # Inherit
        super().__init__(
            context,
            config_path,
            name,
            native_types or self._delegate.natives,
            table_mapping=table_mapping,
            class_types=class_types,
        )

        # Since we've been created with parameters, ensure our config is populated likewise
        self.config["isf_url"] = isf_url
        self.config["symbol_mask"] = symbol_mask

    @staticmethod
    def _closest_version(
        version: str, versions: Dict[Tuple[int, int, int], Type["ISFormatTable"]]
    ) -> Type["ISFormatTable"]:
        """Determines the highest suitable handler for specified version
        format.

        An interface version such as Major.Minor.Patch means that Major
        of the provider must be equal to that of the consumer, and the
        provider (the JSON in this instance) must have a greater minor
        (indicating that only additive changes have been made) than
        the consumer (in this case, the file reader).
        """
        major, minor, patch = (int(x) for x in version.split("."))
        supported_versions = [x for x in versions if x[0] == major and x[1] >= minor]
        if not supported_versions:
            raise ValueError(
                f"No Intermediate Format interface versions support file interface version: {version}"
            )
        return versions[max(supported_versions)]

    symbols = _construct_delegate_function("symbols", True)
    types = _construct_delegate_function("types", True)
    enumerations = _construct_delegate_function("enumerations", True)
    metadata = _construct_delegate_function("metadata", True)
    producer = _construct_delegate_function("producer", True)
    clear_symbol_cache = _construct_delegate_function("clear_symbol_cache")
    get_type = _construct_delegate_function("get_type")
    get_symbol = _construct_delegate_function("get_symbol")
    get_enumeration = _construct_delegate_function("get_enumeration")
    get_type_class = _construct_delegate_function("get_type_class")
    set_type_class = _construct_delegate_function("set_type_class")
    del_type_class = _construct_delegate_function("del_type_class")

    @classmethod
    def file_symbol_url(
        cls, sub_path: str, filename: Optional[str] = None
    ) -> Generator[str, None, None]:
        """Returns an iterator of appropriate file-scheme symbol URLs that can
        be opened by a ResourceAccessor class.

        Filter reduces the number of results returned to only those URLs
        containing that string
        """

        # Check user-modifiable files first, then compressed ones
        extensions = constants.ISF_EXTENSIONS
        if filename is None:
            filename = "*"
            zip_match = filename
        else:
            # For zipfiles, the path separator is always "/", so we need to change the path
            zip_match = "/".join(os.path.split(filename))

        # Check user symbol directory first, then fallback to the framework's library to allow for overloading
        vollog.log(
            constants.LOGLEVEL_VVVV,
            f"Searching for symbols in {', '.join(symbols.__path__)}",
        )
        for path in symbols.__path__:
            if not os.path.isabs(path):
                path = os.path.abspath(os.path.join(__file__, path))
            for extension in extensions:
                # Hopefully these will not be large lists, otherwise this might be slow
                try:
                    for found in (
                        pathlib.Path(path)
                        .joinpath(sub_path)
                        .resolve()
                        .rglob(filename + extension)
                    ):
                        yield found.as_uri()
                except FileNotFoundError:
                    # If there's no linux symbols, don't cry about it
                    pass

            # Finally try looking in zip files
            for zip_path in (
                pathlib.Path(path).joinpath(sub_path).resolve().rglob(filename + ".zip")
            ):
                # We have a zipfile, so run through it and look for sub files that match the filename
                with zipfile.ZipFile(zip_path) as zfile:
                    for name in zfile.namelist():
                        for extension in extensions:
                            # By ending with an extension (and therefore, not /), we should not return any directories
                            if name.endswith(zip_match + extension) or (
                                zip_match == "*" and name.endswith(extension)
                            ):
                                yield (
                                    "jar:file:"
                                    + str(pathlib.Path(zip_path))
                                    + "!"
                                    + name
                                )

    @classmethod
    def create(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        sub_path: str,
        filename: str,
        native_types: Optional[interfaces.symbols.NativeTableInterface] = None,
        table_mapping: Optional[Dict[str, str]] = None,
        class_types: Optional[
            Mapping[str, Type[interfaces.objects.ObjectInterface]]
        ] = None,
        symbol_mask: int = 0,
    ) -> str:
        """Takes a context and loads an intermediate symbol table based on a
        filename.

        Args:
            context: The context that the current plugin is being run within
            config_path: The configuration path for reading/storing configuration information this symbol table may use
            sub_path: The path under a suitable symbol path (defaults to volatility3/symbols and volatility3/framework/symbols) to check
            filename: Basename of the file to find under the sub_path
            native_types: Set of native types, defaults to native types read from the intermediate symbol format file
            table_mapping: a dictionary of table names mentioned within the ISF file, and the tables within the context which they map to
            symbol_mask: An address mask used for all returned symbol offsets from this table (a mask of 0 disables masking)

        Returns:
             the name of the added symbol table
        """
        urls = list(cls.file_symbol_url(sub_path, filename))
        if not urls:
            raise FileNotFoundError(
                f"No symbol files found at provided filename: {filename}",
            )
        table_name = context.symbol_space.free_table_name(filename)
        table = cls(
            context=context,
            config_path=config_path,
            name=table_name,
            isf_url=urls[0],
            native_types=native_types,
            table_mapping=table_mapping,
            class_types=class_types,
            symbol_mask=symbol_mask,
        )
        context.symbol_space.append(table)
        return table_name

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return super().get_requirements() + [
            requirements.StringRequirement(
                "isf_url",
                description="JSON file containing the symbols encoded in the Intermediate Symbol Format",
            ),
        ]


class ISFormatTable(interfaces.symbols.SymbolTableInterface, metaclass=ABCMeta):
    """Provide a base class to identify all subclasses."""

    version = (0, 0, 0)

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        name: str,
        json_object: Any,
        native_types: Optional[interfaces.symbols.NativeTableInterface] = None,
        table_mapping: Optional[Dict[str, str]] = None,
    ) -> None:
        self._json_object = json_object
        self._validate_json()
        self.name = name
        nt = native_types or self._get_natives()
        if nt is None:
            raise TypeError("Native table not provided")
        nt.name = name + "_natives"
        super().__init__(context, config_path, name, nt, table_mapping=table_mapping)
        self._overrides: Dict[str, Type[interfaces.objects.ObjectInterface]] = {}
        self._symbol_cache: Dict[str, interfaces.symbols.SymbolInterface] = {}

    def _get_natives(self) -> Optional[interfaces.symbols.NativeTableInterface]:
        """Determines the appropriate native_types to use from the JSON
        data."""
        # TODO: Consider how to generate the natives entirely from the ISF
        classes = {"x64": native.x64NativeTable, "x86": native.x86NativeTable}
        for nc in sorted(classes):
            native_class = classes[nc]
            for base_type in self._json_object["base_types"]:
                try:
                    if (
                        self._json_object["base_types"][base_type]["length"]
                        != native_class.get_type(base_type).size
                    ):
                        break
                except TypeError:
                    # TODO: determine whether we should give voids a size - We don't give voids a length, whereas microsoft seemingly do
                    pass
            else:
                vollog.debug(f"Choosing appropriate natives for symbol library: {nc}")
                return native_class.natives
        return None

    # TODO: Check the format and make use of the other metadata

    def _validate_json(self) -> None:
        if (
            "user_types" not in self._json_object
            or "base_types" not in self._json_object
            or "metadata" not in self._json_object
            or "symbols" not in self._json_object
            or "enums" not in self._json_object
        ):
            raise exceptions.SymbolSpaceError("Malformed JSON file provided")

    @property
    def metadata(self) -> Optional[interfaces.symbols.MetadataInterface]:
        """Returns a metadata object containing information about the symbol
        table."""
        return None

    @property
    def producer(self) -> Optional["metadata.ProducerMetadata"]:
        """Returns a metadata object containing information about the symbol
        table."""
        return metadata.ProducerMetadata(
            self._json_object.get("metadata", {}).get("producer", {})
        )

    def clear_symbol_cache(self) -> None:
        """Clears the symbol cache of the symbol table."""
        self._symbol_cache.clear()


class Version1Format(ISFormatTable):
    """Class for storing intermediate debugging data as objects and classes."""

    version = (0, 0, 1)

    def get_symbol(self, name: str) -> interfaces.symbols.SymbolInterface:
        """Returns the location offset given by the symbol name."""
        # TODO: Add the ability to add/remove/change symbols after creation
        # note that this should invalidate/update the cache
        if self._symbol_cache.get(name, None):
            return self._symbol_cache[name]
        symbol = self._json_object["symbols"].get(name, None)
        if not symbol:
            raise exceptions.SymbolError(name, self.name, f"Unknown symbol: {name}")
        address = symbol["address"]
        if self.config.get("symbol_mask", 0):
            address = address & self.config["symbol_mask"]

        self._symbol_cache[name] = interfaces.symbols.SymbolInterface(
            name=name, address=address
        )
        return self._symbol_cache[name]

    @property
    def symbols(self) -> Iterable[str]:
        """Returns an iterable (KeysView) of the available symbol names."""
        return self._json_object.get("symbols", {}).keys()

    @property
    def enumerations(self) -> Iterable[Any]:
        """Returns an iterable (KeysView) of the available enumerations."""
        return self._json_object.get("enums", {}).keys()

    @property
    def types(self) -> Iterable[str]:
        """Returns an iterable (KeysView) of the available symbol type names."""
        # We use ** instead of
        # `set(self._json_object.get("user_types", {}).keys()).union(self.natives.types)`
        # because converting user_types dict to a set is costly.
        # It is more efficient to convert the (very small) self.natives.types set to a dict.
        # FIXME: On Python3.8 support drop, merge the two dicts using the merge operator:
        # (self._json_object.get("user_types", {}) | dict.fromkeys(self.natives.types)).keys()
        return {
            **self._json_object.get("user_types", {}),
            **dict.fromkeys(self.natives.types),
        }.keys()

    def get_type_class(self, name: str) -> Type[interfaces.objects.ObjectInterface]:
        return self._overrides.get(name, objects.AggregateType)

    def set_type_class(
        self, name: str, clazz: Type[interfaces.objects.ObjectInterface]
    ) -> None:
        if name not in self.types:
            raise ValueError(f"Symbol type not in {self.name} SymbolTable: {name}")
        self._overrides[name] = clazz

    def del_type_class(self, name: str) -> None:
        if name in self._overrides:
            del self._overrides[name]

    def _interdict_to_template(
        self, dictionary: Dict[str, Any]
    ) -> interfaces.objects.Template:
        """Converts an intermediate format dict into an object template."""
        if not dictionary:
            raise exceptions.SymbolSpaceError(
                f"Invalid intermediate dictionary: {dictionary}"
            )

        type_name = dictionary["kind"]
        if type_name == "base":
            type_name = dictionary["name"]

        if type_name in self.natives.types:
            # The symbol is a native type
            native_template = self.natives.get_type(
                self.name + constants.BANG + type_name
            )

            # Add specific additional parameters, etc
            update = {}
            if type_name == "array":
                update["count"] = dictionary["count"]
                update["subtype"] = self._interdict_to_template(dictionary["subtype"])
            elif type_name == "pointer":
                if dictionary.get("base", None):
                    base_type = self.natives.get_type(
                        self.name + constants.BANG + dictionary["base"]
                    )
                    update["data_format"] = base_type.vol["data_format"]
                update["subtype"] = self._interdict_to_template(dictionary["subtype"])
            elif type_name == "enum":
                update = self._lookup_enum(dictionary["name"])
            elif type_name == "bitfield":
                update = {
                    "start_bit": dictionary["bit_position"],
                    "end_bit": dictionary["bit_position"] + dictionary["bit_length"],
                }
                update["base_type"] = self._interdict_to_template(dictionary["type"])
            # We do *not* call native_template.clone(), since it slows everything down a lot
            # We require that the native.get_type method always returns a newly constructed python object
            native_template.update_vol(**update)
            return native_template

        # Otherwise
        if dictionary["kind"] not in objects.AggregateTypes.values():
            raise exceptions.SymbolSpaceError(
                f"Unknown Intermediate format: {dictionary}"
            )

        reference_name = dictionary["name"]
        if constants.BANG not in reference_name:
            reference_name = self.name + constants.BANG + reference_name
        else:
            reference_parts = reference_name.split(constants.BANG)
            reference_name = (
                self.table_mapping.get(reference_parts[0], reference_parts[0])
                + constants.BANG
                + constants.BANG.join(reference_parts[1:])
            )

        return objects.templates.ReferenceTemplate(type_name=reference_name)

    def _lookup_enum(self, name: str) -> Dict[str, Any]:
        """Looks up an enumeration and returns a dictionary of __init__
        parameters for an Enum."""
        lookup = self._json_object["enums"].get(name, None)
        if not lookup:
            raise exceptions.SymbolSpaceError(f"Unknown enumeration: {name}")
        result = {
            "choices": copy.deepcopy(lookup["constants"]),
            "base_type": self.natives.get_type(lookup["base"]),
        }
        return result

    def get_enumeration(self, enum_name: str) -> interfaces.objects.Template:
        """Resolves an individual enumeration."""
        if constants.BANG in enum_name:
            raise exceptions.SymbolError(
                enum_name,
                self.name,
                f"Enumeration for a different table requested: {enum_name}",
            )
        if enum_name not in self._json_object["enums"]:
            # Fall back to the natives table
            raise exceptions.SymbolError(
                enum_name,
                self.name,
                f"Enumeration not found in {self.name} table: {enum_name}",
            )
        curdict = self._json_object["enums"][enum_name]
        base_type = self.natives.get_type(curdict["base"])
        # The size isn't actually used, the base-type defines it.
        return objects.templates.ObjectTemplate(
            type_name=self.name + constants.BANG + enum_name,
            object_class=objects.Enumeration,
            base_type=base_type,
            choices=curdict["constants"],
        )

    def get_type(self, type_name: str) -> interfaces.objects.Template:
        """Resolves an individual symbol."""
        if constants.BANG in type_name:
            index = type_name.find(constants.BANG)
            table_name, type_name = type_name[:index], type_name[index + 1 :]
            raise exceptions.SymbolError(
                type_name,
                table_name,
                f"Symbol for a different table requested: {table_name + constants.BANG + type_name}",
            )
        if type_name not in self._json_object["user_types"]:
            # Fall back to the natives table
            return self.natives.get_type(self.name + constants.BANG + type_name)
        curdict = self._json_object["user_types"][type_name]
        members = {}
        for member_name in curdict["fields"]:
            interdict = curdict["fields"][member_name]
            member = (
                interdict["offset"],
                self._interdict_to_template(interdict["type"]),
            )
            members[member_name] = member
        object_class = self.get_type_class(type_name)
        if object_class == objects.AggregateType:
            for clazz in objects.AggregateTypes:
                if objects.AggregateTypes[clazz] == curdict["kind"]:
                    object_class = clazz
        return objects.templates.ObjectTemplate(
            type_name=self.name + constants.BANG + type_name,
            object_class=object_class,
            size=curdict["length"],
            members=members,
        )


class Version2Format(Version1Format):
    """Class for storing intermediate debugging data as objects and classes."""

    version = (2, 0, 0)

    def _get_natives(self) -> Optional[interfaces.symbols.NativeTableInterface]:
        """Determines the appropriate native_types to use from the JSON
        data."""
        classes = {"x64": native.x64NativeTable, "x86": native.x86NativeTable}
        for nc in sorted(classes):
            native_class = classes[nc]
            for base_type in self._json_object["base_types"]:
                try:
                    if (
                        self._json_object["base_types"][base_type]["size"]
                        != native_class.get_type(base_type).size
                    ):
                        break
                except TypeError:
                    # TODO: determine whether we should give voids a size - We don't give voids a length, whereas microsoft seemingly do
                    pass
            else:
                vollog.debug(f"Choosing appropriate natives for symbol library: {nc}")
                return native_class.natives
        return None

    def get_type(self, type_name: str) -> interfaces.objects.Template:
        """Resolves an individual symbol."""
        if constants.BANG in type_name:
            index = type_name.find(constants.BANG)
            table_name, type_name = type_name[:index], type_name[index + 1 :]
            raise exceptions.SymbolError(
                type_name,
                table_name,
                f"Symbol for a different table requested: {table_name + constants.BANG + type_name}",
            )
        if type_name not in self._json_object["user_types"]:
            # Fall back to the natives table
            if type_name in self.natives.types:
                return self.natives.get_type(self.name + constants.BANG + type_name)
            else:
                raise exceptions.SymbolError(
                    type_name, self.name, f"Unknown symbol: {type_name}"
                )
        curdict = self._json_object["user_types"][type_name]
        members = {}
        for member_name in curdict["fields"]:
            interdict = curdict["fields"][member_name]
            member = (
                interdict["offset"],
                self._interdict_to_template(interdict["type"]),
            )
            members[member_name] = member
        object_class = self.get_type_class(type_name)
        if object_class == objects.AggregateType:
            for clazz in objects.AggregateTypes:
                if objects.AggregateTypes[clazz] == curdict["kind"]:
                    object_class = clazz
        return objects.templates.ObjectTemplate(
            type_name=self.name + constants.BANG + type_name,
            object_class=object_class,
            size=curdict["size"],
            members=members,
        )


class Version3Format(Version2Format):
    """Class for storing intermediate debugging data as objects and classes."""

    version = (2, 1, 0)

    def get_symbol(self, name: str) -> interfaces.symbols.SymbolInterface:
        """Returns the symbol given by the symbol name."""
        if self._symbol_cache.get(name, None):
            return self._symbol_cache[name]
        symbol = self._json_object["symbols"].get(name, None)
        if not symbol:
            raise exceptions.SymbolError(name, self.name, f"Unknown symbol: {name}")
        address = symbol["address"]
        if self.config.get("symbol_mask", 0):
            address = address & self.config["symbol_mask"]

        symbol_type = None
        if "type" in symbol:
            symbol_type = self._interdict_to_template(symbol["type"])

        self._symbol_cache[name] = interfaces.symbols.SymbolInterface(
            name=name, address=address, type=symbol_type
        )
        return self._symbol_cache[name]


class Version4Format(Version3Format):
    """Class for storing intermediate debugging data as objects and classes."""

    version = (4, 0, 0)

    format_mapping = {
        "int": objects.Integer,
        "float": objects.Float,
        "void": objects.Integer,
        "bool": objects.Boolean,
        "char": objects.Char,
    }

    def _get_natives(self) -> Optional[interfaces.symbols.NativeTableInterface]:
        """Determines the appropriate native_types to use from the JSON
        data."""
        native_dict = {}
        base_types = self._json_object["base_types"]
        for base_type in base_types:
            # Void are ignored because voids are not a volatility primitive, they are a specific Volatility object
            if base_type != "void":
                current = base_types[base_type]
                # TODO: Fix up the typing of this, it bugs out because of the tuple assignment
                if current["kind"] not in self.format_mapping:
                    raise ValueError("Unsupported base kind")
                format_val = (current["size"], current["endian"], current["signed"])
                object_type = self.format_mapping[current["kind"]]
                if base_type == "pointer":
                    object_type = objects.Pointer
                native_dict[base_type] = (object_type, format_val)
        return native.NativeTable(name="native", native_dictionary=native_dict)


class Version5Format(Version4Format):
    """Class for storing intermediate debugging data as objects and classes."""

    version = (4, 1, 0)

    def get_symbol(self, name: str) -> interfaces.symbols.SymbolInterface:
        """Returns the symbol given by the symbol name."""
        if self._symbol_cache.get(name, None):
            return self._symbol_cache[name]
        symbol = self._json_object["symbols"].get(name, None)
        if not symbol:
            raise exceptions.SymbolError(name, self.name, f"Unknown symbol: {name}")
        address = symbol["address"]
        if self.config.get("symbol_mask", 0):
            address = address & self.config["symbol_mask"]

        symbol_type = None
        if "type" in symbol:
            symbol_type = self._interdict_to_template(symbol["type"])
        symbol_constant_data = None
        if "constant_data" in symbol:
            symbol_constant_data = base64.b64decode(symbol.get("constant_data"))

        self._symbol_cache[name] = interfaces.symbols.SymbolInterface(
            name=name,
            address=address,
            type=symbol_type,
            constant_data=symbol_constant_data,
        )
        return self._symbol_cache[name]


class Version6Format(Version5Format):
    """Class for storing intermediate debugging data as objects and classes."""

    version = (6, 0, 0)

    @property
    def metadata(self) -> Optional[interfaces.symbols.MetadataInterface]:
        """Returns a MetadataInterface object."""
        if "metadata" not in self._json_object:
            return None

        json_metadata = self._json_object["metadata"]
        if "windows" in json_metadata:
            return metadata.WindowsMetadata(json_metadata["windows"])
        if "linux" in json_metadata:
            return metadata.LinuxMetadata(json_metadata["linux"])
        if "mac" in json_metadata:
            return metadata.MacMetadata(json_metadata["mac"])

        return None


class Version7Format(Version6Format):
    """Class for storing intermediate debugging data as objects and classes."""

    version = (6, 1, 0)


class Version8Format(Version7Format):
    """Class for storing intermediate debugging data as objects and classes."""

    version = (6, 2, 0)

    def _process_fields(
        self, fields: Dict[str, Dict[str, Any]]
    ) -> Dict[Any, Tuple[int, interfaces.objects.Template]]:
        """For each type field, it walks its tree of subtypes, reducing the hierarchy to just one level.
        It creates a tuple of offset and object templates for each field.
        """
        members = {}
        for new_offset, member_name, member_value in self._reduce_fields(fields):
            member = (new_offset, self._interdict_to_template(member_value["type"]))
            members[member_name] = member
        return members

    def _reduce_fields(
        self, fields: Dict[str, Dict[str, Any]], parent_offset: int = 0
    ) -> Generator[Tuple[int, str, Dict], None, None]:
        """Reduce the fields bringing them one level up. It supports anonymous types such as structs or unions in any
        level of depth."""
        for member_name, member_value in fields.items():
            new_offset = parent_offset + member_value.get("offset", 0)
            if member_value.get("anonymous", False) and isinstance(member_value, dict):
                # Gets the subtype from the json ISF and recursively reduce its fields
                subtype = self._json_object["user_types"].get(
                    member_value["type"]["name"], {}
                )
                yield from self._reduce_fields(subtype["fields"], new_offset)
            else:
                yield new_offset, member_name, member_value

    def get_type(self, type_name: str) -> interfaces.objects.Template:
        """Resolves an individual symbol."""
        index = type_name.find(constants.BANG)
        if index != -1:
            table_name, type_name = type_name[:index], type_name[index + 1 :]
            raise exceptions.SymbolError(
                type_name,
                table_name,
                f"Symbol for a different table requested: {table_name + constants.BANG + type_name}",
            )

        type_definition = self._json_object["user_types"].get(type_name)
        if type_definition is None:
            # Fall back to the natives table
            if type_name in self.natives.types:
                return self.natives.get_type(self.name + constants.BANG + type_name)
            else:
                raise exceptions.SymbolError(
                    type_name, self.name, f"Unknown symbol: {type_name}"
                )

        members = self._process_fields(type_definition["fields"])

        object_class = self.get_type_class(type_name)
        if object_class == objects.AggregateType:
            for clazz in objects.AggregateTypes:
                if objects.AggregateTypes[clazz] == type_definition["kind"]:
                    object_class = clazz
        return objects.templates.ObjectTemplate(
            type_name=self.name + constants.BANG + type_name,
            object_class=object_class,
            size=type_definition["size"],
            members=members,
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
from volatility3.framework import exceptions, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.symbols import mac


class Ifconfig(plugins.PluginInterface):
    """Lists network interface information for all devices"""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="macutils", component=mac.MacUtilities, version=(1, 0, 0)
            ),
        ]

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]

        try:
            list_head = kernel.object_from_symbol(symbol_name="ifnet_head")
        except exceptions.SymbolError:
            list_head = kernel.object_from_symbol(symbol_name="dlil_ifnet_head")

        for ifnet in mac.MacUtilities.walk_tailq(list_head, "if_link"):
            name = utility.pointer_to_string(ifnet.if_name, 32)
            unit = ifnet.if_unit
            prom = ifnet.if_flags & 0x100 == 0x100  # IFF_PROMISC

            sock_addr_dl = ifnet.sockaddr_dl()
            if sock_addr_dl is None:
                mac_addr = renderers.UnreadableValue()
            else:
                mac_addr = str(sock_addr_dl)

            for ifaddr in mac.MacUtilities.walk_tailq(ifnet.if_addrhead, "ifa_link"):
                ip = ifaddr.ifa_addr.get_address()

                yield (0, (f"{name}{unit}", ip, mac_addr, prom))

    def run(self):
        return renderers.TreeGrid(
            [
                ("Interface", str),
                ("IP Address", str),
                ("Mac Address", str),
                ("Promiscuous", bool),
            ],
            self._generator(),
        )



# This file is Copyright 2023 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List

from volatility3.framework import renderers, interfaces, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints

vollog = logging.getLogger(__name__)


class IOMem(interfaces.plugins.PluginInterface):
    """Generates an output similar to /proc/iomem on a running system."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 1)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            )
        ]

    @classmethod
    def parse_resource(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
        resource_offset: int,
        seen: set = set(),
        depth: int = 0,
    ):
        """Recursively parse from a root resource to find details about all related resources.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            vmlinux_module_name: The name of the kernel module on which to operate
            resource_offset: The offset to the resource to be parsed
            seen: The set of resource offsets that have already been parsed
            depth: How deep into the resource structure we are

        Yields:
            Each row of output
        """
        vmlinux = context.modules[vmlinux_module_name]

        # create the resource object with protection against memory smear
        try:
            resource = vmlinux.object("resource", resource_offset, absolute=True)
        except exceptions.InvalidAddressException:
            vollog.warning(
                f"Unable to create resource object at {resource_offset:#x}. This resource, "
                "its sibling, and any of it's children and will be missing from the output."
            )
            return

        # get name with protection against smear as following a pointer
        try:
            name = utility.pointer_to_string(resource.name, 128)
        except exceptions.InvalidAddressException:
            vollog.warning(
                f"Unable to follow pointer to name for resource object at {resource_offset:#x}, "
                "replaced with UnreadableValue"
            )
            name = renderers.UnreadableValue()

        try:
            start = resource.start
            end = resource.end
        except exceptions.InvalidAddressException:
            vollog.warning(
                f"Unable to follow pointer to start and end for resource object at {resource_offset:#x}. Skipping entry."
            )
            return

        # mark this resource as seen in the seen set. Normally this should not be needed but will protect
        # against possible infinite loops. Warn the user if an infinite loop would have happened.
        if resource_offset in seen:
            vollog.warning(
                f"The resource object at {resource_offset:#x} '{name}' has already been processed, "
                "this should not normally occur. No further results from related resources will be "
                "displayed to protect against infinite loops."
            )
            return
        else:
            seen.add(resource_offset)

        # yield information on this resource
        yield depth, (name, start, end)

        # process child resource if this exists
        if resource.child != 0:
            yield from cls.parse_resource(
                context,
                vmlinux_module_name,
                resource.child,
                seen,
                depth + 1,
            )

        # process sibling resource if this exists
        if resource.sibling != 0:
            yield from cls.parse_resource(
                context,
                vmlinux_module_name,
                resource.sibling,
                seen,
                depth,
            )

    def _generator(self):
        """Generates an output similar to /proc/iomem on a running system

        Args:
            None

        Yields:
            Each row of output using the parse_resource function
        """

        # get the kernel module from the current context
        vmlinux_module_name = self.config["kernel"]
        vmlinux = self.context.modules[vmlinux_module_name]

        # get the address for the iomem_resource
        try:
            iomem_root_offset = vmlinux.get_absolute_symbol_address("iomem_resource")
        except exceptions.SymbolError:
            iomem_root_offset = None

        # only continue if iomem_root address was located
        if iomem_root_offset is not None:
            # recursively parse the resources starting from the root resource at 'iomem_resource'
            for depth, (name, start, end) in self.parse_resource(
                self.context, vmlinux_module_name, iomem_root_offset
            ):
                # use format_hints to format start and end addresses for the renderers
                yield depth, (name, format_hints.Hex(start), format_hints.Hex(end))

    def run(self):
        # get the kernel module from the current context
        vmlinux_module_name = self.config["kernel"]
        vmlinux = self.context.modules[vmlinux_module_name]

        # check that the iomem_resource symbol exists
        # normally exported in /kernel/resource.c
        if not vmlinux.has_symbol("iomem_resource"):
            raise TypeError(
                "This plugin requires the iomem_resource symbol. This symbol is not present in the supplied symbol table. This means you are either analyzing an unsupported kernel version or that your symbol table is corrupt."
            )

        # error if type 'resource' is not found
        if not vmlinux.has_type("resource"):
            raise TypeError(
                "This plugin requires the resource type. This type is not present in the supplied symbol table. This means you are either analyzing an unsupported kernel version or that your symbol table is corrupt."
            )

        columns = [
            ("Name", str),
            ("Start", format_hints.Hex),
            ("End", format_hints.Hex),
        ]
        return renderers.TreeGrid(columns, self._generator())



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import math
import logging
import datetime
import time
import tarfile
from dataclasses import dataclass, astuple
from typing import IO, List, Set, Type, Iterable, Tuple, Union
from io import BytesIO
from pathlib import PurePath

from volatility3.framework.constants import architectures
from volatility3.framework import constants, renderers, interfaces, exceptions
from volatility3.framework.renderers import format_hints
from volatility3.framework.interfaces import plugins
from volatility3.framework.configuration import requirements
from volatility3.plugins import timeliner
from volatility3.plugins.linux import mountinfo

vollog = logging.getLogger(__name__)


@dataclass
class InodeUser:
    """Inode user representation, featuring augmented information and formatted fields.
    This is the data the plugin will eventually display.
    """

    superblock_addr: int
    mountpoint: str
    device: str
    inode_num: int
    inode_addr: int
    type: str
    inode_pages: int
    cached_pages: int
    file_mode: str
    access_time: str
    modification_time: str
    change_time: str
    path: str
    inode_size: int

    @classmethod
    def format_symlink(cls, symlink_source: str, symlink_dest: str) -> str:
        return f"{symlink_source} -> {symlink_dest}"


@dataclass
class InodeInternal:
    """Inode internal representation containing only the core objects

    Fields:
        superblock: 'super_block' struct
        mountpoint: Superblock mountpoint path
        inode: 'inode' struct
        path: Dentry full path
    """

    superblock: interfaces.objects.ObjectInterface
    mountpoint: str
    inode: interfaces.objects.ObjectInterface
    path: str

    def to_user(
        self, kernel_layer: interfaces.layers.TranslationLayerInterface
    ) -> InodeUser:
        """Augment the inode information to be presented to the user

        Args:
            kernel_layer: The kernel layer to obtain the page size

        Returns:
            An InodeUser dataclass
        """
        # Ensure all types are atomic immutable. Otherwise, astuple() will take a long
        # time doing a deepcopy of the Volatility objects.
        superblock_addr = self.superblock.vol.offset
        device = f"{self.superblock.major}:{self.superblock.minor}"
        inode_num = int(self.inode.i_ino)
        inode_addr = self.inode.vol.offset
        inode_type = self.inode.get_inode_type() or renderers.UnparsableValue()
        # Round up the number of pages to fit the inode's size
        inode_pages = int(math.ceil(self.inode.i_size / float(kernel_layer.page_size)))
        cached_pages = int(self.inode.i_mapping.nrpages)
        file_mode = self.inode.get_file_mode()
        access_time_dt = self.inode.get_access_time()
        modification_time_dt = self.inode.get_modification_time()
        change_time_dt = self.inode.get_change_time()
        inode_size = int(self.inode.i_size)

        inode_user = InodeUser(
            superblock_addr=superblock_addr,
            mountpoint=self.mountpoint,
            device=device,
            inode_num=inode_num,
            inode_addr=inode_addr,
            type=inode_type,
            inode_pages=inode_pages,
            cached_pages=cached_pages,
            file_mode=file_mode,
            access_time=access_time_dt,
            modification_time=modification_time_dt,
            change_time=change_time_dt,
            path=self.path,
            inode_size=inode_size,
        )
        return inode_user


class Files(plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Lists files from memory"""

    _required_framework_version = (2, 0, 0)

    _version = (1, 1, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="mountinfo", component=mountinfo.MountInfo, version=(1, 2, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.ListRequirement(
                name="type",
                description="List of space-separated file type filters i.e. --type REG DIR",
                element_type=str,
                optional=True,
            ),
            requirements.StringRequirement(
                name="find",
                description="Filename (full path) to find",
                optional=True,
            ),
        ]

    @staticmethod
    def _follow_symlink(
        inode: interfaces.objects.ObjectInterface,
        symlink_path: str,
    ) -> str:
        """Follows (fast) symlinks (kernels >= 4.2.x).
        Fast symlinks are filesystem agnostic.

        Args:
            inode: The inode (or pointer) to dump
            symlink_path: The symlink name

        Returns:
            If it can resolve the symlink, it returns a string "symlink_path -> target_path"
            Otherwise, it returns the same symlink_path
        """
        # i_link (fast symlinks) were introduced in 4.2
        if (
            inode
            and inode.is_link
            and inode.has_member("i_link")
            and inode.i_link
            and inode.i_link.is_readable()
        ):
            symlink_dest = inode.i_link.dereference().cast(
                "string", max_length=255, encoding="utf-8", errors="replace"
            )
            symlink_path = InodeUser.format_symlink(symlink_path, symlink_dest)

        return symlink_path

    @classmethod
    def _walk_dentry(
        cls,
        seen_dentries: Set[int],
        root_dentry: interfaces.objects.ObjectInterface,
        parent_dir: str,
    ):
        """Walks dentries recursively

        Args:
            seen_dentries: A set to ensure each dentry is processed only once
            root_dentry: Root dentry object
            parent_dir: Parent directory path

        Yields:
           file_path: Filename including path
           dentry: Dentry object
        """

        for dentry in root_dentry.get_subdirs():
            dentry_addr = dentry.vol.offset

            # corruption
            if dentry_addr == root_dentry.vol.offset:
                continue

            if dentry_addr in seen_dentries:
                continue

            seen_dentries.add(dentry_addr)

            inode_ptr = dentry.d_inode
            if not (inode_ptr and inode_ptr.is_readable()):
                continue

            inode = inode_ptr.dereference()
            if not inode.is_valid():
                continue

            # This allows us to have consistent paths
            if dentry.d_name.name:
                basename = dentry.d_name.name_as_str()
                # Do NOT use os.path.join() below
                file_path = parent_dir + "/" + basename
            else:
                continue

            yield file_path, dentry

            if inode.is_dir:
                yield from cls._walk_dentry(seen_dentries, dentry, parent_dir=file_path)

    @classmethod
    def get_inodes(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
        follow_symlinks: bool = True,
    ) -> Iterable[InodeInternal]:
        """Retrieves the inodes from the superblocks

        Args:
            context: The context that the plugin will operate within
            vmlinux_module_name: The name of the kernel module on which to operate
            follow_symlinks: Whether to follow symlinks or not

        Yields:
            An InodeInternal object
        """

        superblocks_iter = mountinfo.MountInfo.get_superblocks(
            context=context,
            vmlinux_module_name=vmlinux_module_name,
        )

        seen_inodes = set()
        seen_dentries = set()
        for superblock, mountpoint in superblocks_iter:
            parent_dir = "" if mountpoint == "/" else mountpoint

            # Superblock root dentry
            root_dentry_ptr = superblock.s_root
            if not root_dentry_ptr:
                continue

            root_dentry = root_dentry_ptr.dereference()

            # Dentry sanity check
            if not root_dentry.is_root():
                continue

            # More dentry/inode sanity checks
            root_inode_ptr = root_dentry.d_inode
            if not (root_inode_ptr and root_inode_ptr.is_readable()):
                continue

            root_inode = root_inode_ptr.dereference()
            if not root_inode.is_valid():
                continue

            if not (root_inode.i_mapping and root_inode.i_mapping.is_readable()):
                # Retrieving data from the page cache requires a valid address space
                continue

            # Inode already processed?
            # Store a primitive int (instead of the pointer value) to track
            # addresses we've already seen. Storing the full `objects.Pointer`
            # uses too much memory, and we don't need all of the information
            # that it contains.
            if root_inode_ptr in seen_inodes:
                continue

            seen_inodes.add(int(root_inode_ptr))

            root_path = mountpoint

            inode_in = InodeInternal(
                superblock=superblock,
                mountpoint=mountpoint,
                inode=root_inode,
                path=root_path,
            )
            yield inode_in

            # Children
            for file_path, file_dentry in cls._walk_dentry(
                seen_dentries, root_dentry, parent_dir
            ):
                if not file_dentry:
                    continue

                # Dentry/inode sanity checks
                file_inode_ptr = file_dentry.d_inode
                if not (file_inode_ptr and file_inode_ptr.is_readable()):
                    continue

                file_inode = file_inode_ptr.dereference()
                if not file_inode.is_valid():
                    continue

                if not (file_inode.i_mapping and file_inode.i_mapping.is_readable()):
                    # Retrieving data from the page cache requires a valid address space
                    continue

                # Inode already processed?
                # Store a primitive int (instead of the pointer value) to track
                # addresses we've already seen. Storing the full `objects.Pointer`
                # uses too much memory, and we don't need all of the information
                # that it contains.
                if file_inode_ptr in seen_inodes:
                    continue
                seen_inodes.add(int(file_inode_ptr))

                if follow_symlinks:
                    file_path = cls._follow_symlink(file_inode_ptr, file_path)
                inode_in = InodeInternal(
                    superblock=superblock,
                    mountpoint=mountpoint,
                    inode=file_inode,
                    path=file_path,
                )
                yield inode_in

    def _generator(self):
        vmlinux_module_name = self.config["kernel"]
        vmlinux = self.context.modules[vmlinux_module_name]
        vmlinux_layer = self.context.layers[vmlinux.layer_name]

        inodes_iter = self.get_inodes(
            context=self.context,
            vmlinux_module_name=vmlinux_module_name,
        )

        types_filter = self.config["type"]
        for inode_in in inodes_iter:
            if types_filter and inode_in.inode.get_inode_type() not in types_filter:
                continue

            if self.config["find"]:
                if inode_in.path == self.config["find"]:
                    inode_out = inode_in.to_user(vmlinux_layer)

                    yield (0, astuple(inode_out))
                    break  # Only the first match
            else:
                inode_out = inode_in.to_user(vmlinux_layer)

                yield (0, astuple(inode_out))

    def generate_timeline(self):
        """Generates tuples of (description, timestamp_type, timestamp)

        These need not be generated in any particular order, sorting
        will be done later
        """
        vmlinux_module_name = self.config["kernel"]
        vmlinux = self.context.modules[vmlinux_module_name]
        vmlinux_layer = self.context.layers[vmlinux.layer_name]

        inodes_iter = self.get_inodes(
            context=self.context,
            vmlinux_module_name=vmlinux_module_name,
        )

        for inode_in in inodes_iter:
            inode_out = inode_in.to_user(vmlinux_layer)
            description = f"Cached Inode for {inode_out.path}"
            yield description, timeliner.TimeLinerType.ACCESSED, inode_out.access_time
            yield (
                description,
                timeliner.TimeLinerType.MODIFIED,
                inode_out.modification_time,
            )
            yield description, timeliner.TimeLinerType.CHANGED, inode_out.change_time

    @classmethod
    def format_fields_with_headers(cls, headers, generator):
        """Uses the headers type to cast the fields obtained from the generator"""
        for level, fields in generator:
            formatted_fields = []
            for header, field in zip(headers, fields):
                header_type = header[1]

                if isinstance(
                    field, (header_type, interfaces.renderers.BaseAbsentValue)
                ):
                    formatted_field = field
                else:
                    formatted_field = header_type(field)

                formatted_fields.append(formatted_field)
            yield level, formatted_fields

    def run(self):
        headers = [
            ("SuperblockAddr", format_hints.Hex),
            ("MountPoint", str),
            ("Device", str),
            ("InodeNum", int),
            ("InodeAddr", format_hints.Hex),
            ("FileType", str),
            ("InodePages", int),
            ("CachedPages", int),
            ("FileMode", str),
            ("AccessTime", datetime.datetime),
            ("ModificationTime", datetime.datetime),
            ("ChangeTime", datetime.datetime),
            ("FilePath", str),
            ("InodeSize", int),
        ]

        return renderers.TreeGrid(
            headers, self.format_fields_with_headers(headers, self._generator())
        )


class InodePages(plugins.PluginInterface):
    """Lists and recovers cached inode pages"""

    _required_framework_version = (2, 0, 0)

    _version = (3, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="files", component=Files, version=(1, 0, 0)
            ),
            requirements.StringRequirement(
                name="find",
                description="Filename (full path) to find ",
                optional=True,
            ),
            requirements.IntRequirement(
                name="inode",
                description="Inode address",
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract inode content",
                default=False,
                optional=True,
            ),
        ]

    @classmethod
    def write_inode_content_to_file(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        inode: interfaces.objects.ObjectInterface,
        filename: str,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
    ) -> None:
        """Extracts the inode's contents from the page cache and saves them to a file

        Args:
            context: The context on which to operate
            layer_name: The name of the layer on which to operate
            inode: The inode to dump
            filename: Filename for writing the inode content
            open_method: class for constructing output files
        """
        try:
            with open_method(filename) as file_obj:
                cls.write_inode_content_to_stream(context, layer_name, inode, file_obj)
        except OSError as e:
            vollog.error("Unable to write to file (%s): %s", filename, e)

    @classmethod
    def write_inode_content_to_stream(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        inode: interfaces.objects.ObjectInterface,
        stream: IO,
    ) -> None:
        """Extracts the inode's contents from the page cache and saves them to a stream

        Args:
            context: The context on which to operate
            layer_name: The name of the layer on which to operate
            inode: The inode to dump
            stream: An IO stream to write to, typically FileHandlerInterface or BytesIO
        """
        if not inode.is_reg:
            vollog.error("The inode is not a regular file")
            return None

        layer = context.layers[layer_name]
        # By using truncate/seek, provided the filesystem supports it, and the
        # stream is a File interface, a sparse file will be
        # created, saving both disk space and I/O time.
        # Additionally, using the page index will guarantee that each page is written at the
        # appropriate file position.
        inode_size = inode.i_size
        try:
            stream_initialized = False
            for page_idx, page_content in inode.get_contents():
                current_fp = page_idx * layer.page_size
                max_length = inode_size - current_fp
                page_bytes_len = min(max_length, len(page_content))
                if current_fp >= inode_size or current_fp + page_bytes_len > inode_size:
                    vollog.error(
                        "Page out of file bounds: inode 0x%x, inode size %d, page index %d",
                        inode.vol.offset,
                        inode_size,
                        page_idx,
                    )
                    continue
                page_bytes = page_content[:page_bytes_len]

                if not stream_initialized:
                    # Lazy initialization to avoid truncating the stream until we are
                    # certain there is something to write
                    stream.truncate(inode_size)
                    stream_initialized = True

                stream.seek(current_fp)
                stream.write(page_bytes)
        except exceptions.LinuxPageCacheException:
            vollog.error(
                f"Error dumping cached pages for inode at {inode.vol.offset:#x}"
            )

    def _generate_inode_fields(
        self,
        inode: interfaces.objects.ObjectInterface,
        vmlinux_layer: interfaces.layers.TranslationLayerInterface,
        filename: Union[renderers.NotApplicableValue, str],
    ) -> Iterable[Tuple[int, int, int, int, bool, str]]:
        inode_size = inode.i_size
        try:
            for page_obj in inode.get_pages():
                if page_obj.mapping != inode.i_mapping:
                    vollog.warning(
                        f"Cached page at {page_obj.vol.offset:#x} has a mismatched address space with the inode. Skipping page"
                    )
                    continue
                page_vaddr = page_obj.vol.offset
                page_paddr = page_obj.to_paddr()
                page_mapping_addr = page_obj.mapping
                page_index = page_obj.index
                page_file_offset = page_index * vmlinux_layer.page_size
                dump_safe = (
                    page_file_offset < inode_size
                    and page_mapping_addr
                    and page_mapping_addr.is_readable()
                )
                page_flags_list = page_obj.get_flags_list()
                page_flags = ",".join([x.replace("PG_", "") for x in page_flags_list])
                fields = (
                    page_vaddr,
                    page_paddr,
                    page_mapping_addr,
                    page_index,
                    dump_safe,
                    page_flags,
                    filename,
                )

                yield 0, fields
        except exceptions.LinuxPageCacheException:
            vollog.warning(f"Page cache for inode at {inode.vol.offset:#x} is corrupt")

    def _generator(self):
        vmlinux_module_name = self.config["kernel"]
        vmlinux = self.context.modules[vmlinux_module_name]
        vmlinux_layer = self.context.layers[vmlinux.layer_name]

        if self.config["inode"] and self.config["find"]:
            vollog.error("Cannot use --inode and --find simultaneously")
            return None

        if self.config["find"]:
            inodes_iter = Files.get_inodes(
                context=self.context,
                vmlinux_module_name=vmlinux_module_name,
            )
            for inode_in in inodes_iter:
                if inode_in.path == self.config["find"]:
                    inode = inode_in.inode
                    break  # Only the first match
            else:
                vollog.error("Unable to find inode with path %s", self.config["find"])
                return None
        elif self.config["inode"]:
            inode = vmlinux.object("inode", self.config["inode"], absolute=True)
        else:
            vollog.error("You must use either --inode or --find")
            return None

        if not inode.is_valid():
            vollog.error("Invalid inode at 0x%x", inode.vol.offset)
            return None

        if not inode.is_reg:
            vollog.error("The inode is not a regular file")
            return None

        filename = renderers.NotApplicableValue()
        if self.config["dump"]:
            open_method = self.open
            inode_address = inode.vol.offset
            filename = open_method.sanitize_filename(f"inode_0x{inode_address:x}.dmp")
            vollog.info("[*] Writing inode at 0x%x to '%s'", inode_address, filename)
            self.write_inode_content_to_file(
                self.context, vmlinux_layer.name, inode, filename, open_method
            )
        yield from self._generate_inode_fields(inode, vmlinux_layer, filename)

    def run(self):
        headers = [
            ("PageVAddr", format_hints.Hex),
            ("PagePAddr", format_hints.Hex),
            ("MappingAddr", format_hints.Hex),
            ("Index", int),
            ("DumpSafe", bool),
            ("Flags", str),
            ("Output File", str),
        ]

        return renderers.TreeGrid(
            headers, Files.format_fields_with_headers(headers, self._generator())
        )


class RecoverFs(plugins.PluginInterface):
    """Recovers the cached filesystem (directories, files, symlinks) into a compressed tarball.

    Details: level 0 directories are named after the UUID of the parent superblock; metadata aren't replicated to extracted objects; objects modification time is set to the plugin run time; absolute symlinks
    are converted to relative symlinks to prevent referencing the analyst's filesystem.
    Troubleshooting: to fix extraction errors related to long paths, please consider using https://github.com/mxmlnkn/ratarmount.
    """

    _version = (1, 0, 1)
    _required_framework_version = (2, 21, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="files", component=Files, version=(1, 1, 0)
            ),
            requirements.VersionRequirement(
                name="inodepages", component=InodePages, version=(3, 0, 0)
            ),
            requirements.BooleanRequirement(
                name="tmpfs_only",
                description="Extracts only files from tmpfs file systems",
                default=False,
                optional=True,
            ),
            requirements.ChoiceRequirement(
                name="compression_format",
                description="Compression format (default: gz)",
                choices=["gz", "bz2", "xz"],
                default="gz",
                optional=True,
            ),
        ]

    def _tar_add_reg_inode(
        self,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        tar: tarfile.TarFile,
        reg_inode_in: InodeInternal,
        path_prefix: str = "",
        mtime: float = None,
    ) -> int:
        """Extracts a REG inode content and writes it to a TarFile object.

        Args:
            context: The context on which to operate
            layer_name: The name of the layer on which to operate
            tar: The TarFile object to write to
            reg_inode_in: The inode to extract content from
            path_prefix: A custom path prefix to prepend the inode path with
            mtime: The modification time to set the TarInfo object to

        Returns:
            The number of extracted bytes
        """
        inode_content_buffer = BytesIO()
        InodePages.write_inode_content_to_stream(
            context, layer_name, reg_inode_in.inode, inode_content_buffer
        )
        inode_content_buffer.seek(0)
        handle_buffer_size = inode_content_buffer.getbuffer().nbytes

        tar_info = tarfile.TarInfo(path_prefix + reg_inode_in.path)
        # The tarfile module only has read support for sparse files:
        # https://docs.python.org/3.12/library/tarfile.html#tarfile.LNKTYPE:~:text=and%20longlink%20extensions%2C-,read%2Donly%20support,-for%20all%20variants
        tar_info.type = tarfile.REGTYPE
        tar_info.size = handle_buffer_size
        tar_info.mode = 0o444
        if mtime is not None:
            tar_info.mtime = mtime
        tar.addfile(tar_info, inode_content_buffer)

        return handle_buffer_size

    def _tar_add_dir(
        self,
        tar: tarfile.TarFile,
        directory_path: str,
        mtime: float = None,
    ) -> None:
        """Adds a directory path to a TarFile object, based on a DIR inode.

        Args:
            tar: The TarFile object to write to
            directory_path: The directory path to create
            mtime: The modification time to set the TarInfo object to
        """
        tar_info = tarfile.TarInfo(directory_path)
        tar_info.type = tarfile.DIRTYPE
        tar_info.mode = 0o755
        if mtime is not None:
            tar_info.mtime = mtime
        tar.addfile(tar_info)

    def _tar_add_lnk(
        self,
        tar: tarfile.TarFile,
        symlink_source: str,
        symlink_dest: str,
        symlink_source_prefix: str = "",
        mtime: float = None,
    ) -> None:
        """Adds a symlink to a TarFile object.

        Args:
            tar: The TarFile object to write to
            symlink_source: The symlink source path
            symlink_dest: The symlink target/destination
            symlink_source_prefix: A custom path prefix to prepend the symlink source with
            mtime: The modification time to set the TarInfo object to
        """
        # Patch symlinks pointing to absolute paths,
        # to prevent referencing the host filesystem.
        if symlink_dest.startswith("/"):
            relative_dest = PurePath(symlink_dest).relative_to(PurePath("/"))
            # Remove the leading "/" to prevent an extra undesired "../" in the output
            symlink_dest = (
                PurePath(
                    *[".."] * len(PurePath(symlink_source.lstrip("/")).parent.parts)
                )
                / relative_dest
            ).as_posix()
        tar_info = tarfile.TarInfo(symlink_source_prefix + symlink_source)
        tar_info.type = tarfile.SYMTYPE
        tar_info.linkname = symlink_dest
        tar_info.mode = 0o444
        if mtime is not None:
            tar_info.mtime = mtime
        tar.addfile(tar_info)

    def _generator(self):
        vmlinux_module_name = self.config["kernel"]
        vmlinux = self.context.modules[vmlinux_module_name]
        vmlinux_layer = self.context.layers[vmlinux.layer_name]
        tar_buffer = BytesIO()
        tar = tarfile.open(
            fileobj=tar_buffer,
            mode=f"w:{self.config['compression_format']}",
        )
        # Set a unique timestamp for all extracted files
        mtime = time.time()

        inodes_iter = Files.get_inodes(
            context=self.context,
            vmlinux_module_name=vmlinux_module_name,
            follow_symlinks=False,
        )

        # Prefix paths with the superblock UUID's to prevent overlaps.
        # Switch to device major and device minor for older kernels (< 2.6.39-rc1).
        uuid_as_prefix = vmlinux.get_type("super_block").has_member("s_uuid")
        if not uuid_as_prefix:
            vollog.warning(
                "super_block struct does not support s_uuid attribute. Consequently, level 0 directories won't refer to the superblock uuid's, but to its device_major:device_minor numbers."
            )

        visited_paths = seen_prefixes = set()
        for inode_in in inodes_iter:
            # Code is slightly duplicated here with the if-block below.
            # However this prevents unneeded tar manipulation if fifo
            # or sock inodes come through for example.
            if not (
                inode_in.inode.is_reg or inode_in.inode.is_dir or inode_in.inode.is_link
            ):
                continue

            if not inode_in.path.startswith("/"):
                vollog.debug(
                    f'Skipping processing of potentially smeared "{inode_in.path}" inode name as it does not starts with a "/".'
                )
                continue

            sb_type = inode_in.superblock.get_type()
            if not sb_type:
                vollog.debug(
                    f"Unable to read superblock type for inode at {inode_in.inode.vol.offset}"
                )
                continue

            if self.config["tmpfs_only"] and sb_type != "tmpfs":
                vollog.debug(f"Skipping non-tmpfs filesystem {sb_type}")
                continue

            # Construct the output path
            if uuid_as_prefix:
                prefix = f"/{inode_in.superblock.uuid}"
            else:
                prefix = f"/{inode_in.superblock.major}:{inode_in.superblock.minor}"
            prefixed_path = prefix + inode_in.path

            # Sanity check for already processed paths
            if prefixed_path in visited_paths:
                vollog.log(
                    constants.LOGLEVEL_VV,
                    f'Already processed prefixed inode path: "{prefixed_path}".',
                )
                continue
            elif prefix not in seen_prefixes:
                self._tar_add_dir(tar, prefix, mtime)
                seen_prefixes.add(prefix)

            visited_paths.add(prefixed_path)
            extracted_file_size = renderers.NotApplicableValue()

            # Inodes parent directory is yielded first, which
            # ensures that a file parent path will exist beforehand.
            # tarfile will take care of creating it anyway.
            if inode_in.inode.is_reg:
                extracted_file_size = self._tar_add_reg_inode(
                    self.context,
                    vmlinux_layer.name,
                    tar,
                    inode_in,
                    prefix,
                    mtime,
                )
            elif inode_in.inode.is_dir:
                self._tar_add_dir(tar, prefixed_path, mtime)
            elif (
                inode_in.inode.is_link
                and inode_in.inode.has_member("i_link")
                and inode_in.inode.i_link
                and inode_in.inode.i_link.is_readable()
            ):
                symlink_dest = inode_in.inode.i_link.dereference().cast(
                    "string", max_length=255, encoding="utf-8", errors="replace"
                )
                self._tar_add_lnk(tar, inode_in.path, symlink_dest, prefix, mtime)
                # Set path to a user friendly representation before yielding
                inode_in.path = InodeUser.format_symlink(inode_in.path, symlink_dest)
            else:
                continue

            inode_out = inode_in.to_user(vmlinux_layer)
            yield (0, astuple(inode_out) + (extracted_file_size,))

        tar.close()
        tar_buffer.seek(0)
        output_filename = f"recovered_fs.tar.{self.config['compression_format']}"
        with self.open(output_filename) as f:
            f.write(tar_buffer.getvalue())

    def run(self):
        headers = [
            ("SuperblockAddr", format_hints.Hex),
            ("MountPoint", str),
            ("Device", str),
            ("InodeNum", int),
            ("InodeAddr", format_hints.Hex),
            ("FileType", str),
            ("InodePages", int),
            ("CachedPages", int),
            ("FileMode", str),
            ("AccessTime", datetime.datetime),
            ("ModificationTime", datetime.datetime),
            ("ChangeTime", datetime.datetime),
            ("FilePath", str),
            ("InodeSize", int),
            ("Recovered FileSize", int),
        ]

        return renderers.TreeGrid(
            headers, Files.format_fields_with_headers(headers, self._generator())
        )
